/**********************************************************/
/* Mach3: This is an altered version of tcp_hdr.c         */
/*        to be used when building an in-kernel version   */
/*        of the x-kernel. That is due to the fact that   */
/*        up to now usc/usit produce strict ANSI C code   */
/*        while in-kernel code is compiled with a K&R cc. */
/*        It works for mach3-mips and mach3-intelx86.     */
/*                                                        */
/* This should be fixed in the next usc/usit release.     */
/*                                                        */
/* Based on the files listed below as USC/USIT input      */
/**********************************************************/

/**********************************************************/
/*                                                        */
/*                  USC Output                            */
/*                  Wed Nov 16 16:04:38 1994              */
/*                  File: tcp_hdr.usc                     */
/*                                                        */
/**********************************************************/

/**********************************************************/
/*                                                        */
/*                  USIT Output                           */
/*                  Wed Nov 16 16:03:59 1994              */
/*                  File: tcp_hdr.usit                    */
/*                                                        */
/**********************************************************/
/* 
 * $File: $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 * usc/usit v1.0
 *
 * Copyright (c) 1994  Arizona Board of Regents
 *
 * $Revision: 1.1 $
 * $Date: 1994/11/23 16:52:39 $
 */
/*
 * TCP header store and load functions
 */
#include "xkernel.h"
#include "tcp_internal.h"
#include "tcp_hdr.h"

#define CKSUM_TRACE 8

void
tcpHdrCopyIn( l_native_header, l_net_header)  
  void* l_native_header;
  void* l_net_header;
{
  unsigned long r0;
  r0 = *(unsigned int *)((unsigned char *)l_net_header + (0));
  *((int *)((char *)l_native_header + (0))) =
      (((r0 >> 8) & 0xff00ffL) |
    ((r0 << 8) & 0xff00ff00L));
  r0 = *(unsigned int *)((unsigned char *)l_net_header + (4));
  *((int *)((char *)l_native_header + (4))) =
      (((r0 >> 24) & 0xffL) |
    ((r0 >> 8) & 0xff00L) |
    ((r0 << 8) & 0xff0000L) |
    ((r0 << 24) & 0xff000000L));
  r0 = *(unsigned int *)((unsigned char *)l_net_header + (8));
  *((int *)((char *)l_native_header + (8))) =
      (((r0 >> 24) & 0xffL) |
    ((r0 >> 8) & 0xff00L) |
    ((r0 << 8) & 0xff0000L) |
    ((r0 << 24) & 0xff000000L));
  r0 = *(unsigned int *)((unsigned char *)l_net_header + (12));
  *((int *)((char *)l_native_header + (12))) =
      (((r0 << 4) & 0xf0L) |
    ((r0 >> 4) & 0xfL) |
    (r0 & 0xff00L) |
    ((r0 >> 8) & 0xff0000L) |
    ((r0 << 8) & 0xff000000L));
  r0 = *(unsigned int *)((unsigned char *)l_net_header + (16));
  *((int *)((char *)l_native_header + (16))) =
      (((r0 >> 8) & 0xff00ffL) |
    ((r0 << 8) & 0xff00ff00L));
}


void 
tcpHdrCopyOut(l_net_header, l_native_header)  
  void* l_net_header;
  void* l_native_header;
{
  unsigned long r0;
  r0 = *(unsigned int *)((unsigned char *)l_native_header + (0));
  *((int *)((char *)l_net_header + (0))) =
      (((r0 >> 8) & 0xff00ffL) |
    ((r0 << 8) & 0xff00ff00L));
  r0 = *(unsigned int *)((unsigned char *)l_native_header + (4));
  *((int *)((char *)l_net_header + (4))) =
      (((r0 >> 24) & 0xffL) |
    ((r0 >> 8) & 0xff00L) |
    ((r0 << 8) & 0xff0000L) |
    ((r0 << 24) & 0xff000000L));
  r0 = *(unsigned int *)((unsigned char *)l_native_header + (8));
  *((int *)((char *)l_net_header + (8))) =
      (((r0 >> 24) & 0xffL) |
    ((r0 >> 8) & 0xff00L) |
    ((r0 << 8) & 0xff0000L) |
    ((r0 << 24) & 0xff000000L));
  r0 = *(unsigned int *)((unsigned char *)l_native_header + (12));
  *((int *)((char *)l_net_header + (12))) =
      (((r0 << 4) & 0xf0L) |
    ((r0 >> 4) & 0xfL) |
    (r0 & 0xff00L) |
    ((r0 >> 8) & 0xff0000L) |
    ((r0 << 8) & 0xff000000L));
  r0 = *(unsigned int *)((unsigned char *)l_native_header + (16));
  *((int *)((char *)l_net_header + (16))) =
      (((r0 >> 8) & 0xff00ffL) |
    ((r0 << 8) & 0xff00ff00L));
}

void
tcpHdrCopyUOut(l_net_header, l_native_header)  
  void* l_net_header;
  void* l_native_header;
{
  unsigned long r0;
  r0 = *(unsigned int *)((unsigned char *)l_native_header + (0));
  *((char *)((char *)l_net_header + (0))) =
      ((r0 >> 8));
  *((char *)((char *)l_net_header + (1))) =
      (r0);
  *((char *)((char *)l_net_header + (2))) =
      ((r0 >> 24));
  *((char *)((char *)l_net_header + (3))) =
      ((r0 >> 16));
  r0 = *(unsigned int *)((unsigned char *)l_native_header + (4));
  *((char *)((char *)l_net_header + (4))) =
      ((r0 >> 24));
  *((char *)((char *)l_net_header + (5))) =
      ((r0 >> 16));
  *((char *)((char *)l_net_header + (6))) =
      ((r0 >> 8));
  *((char *)((char *)l_net_header + (7))) =
      (r0);
  r0 = *(unsigned int *)((unsigned char *)l_native_header + (8));
  *((char *)((char *)l_net_header + (8))) =
      ((r0 >> 24));
  *((char *)((char *)l_net_header + (9))) =
      ((r0 >> 16));
  *((char *)((char *)l_net_header + (10))) =
      ((r0 >> 8));
  *((char *)((char *)l_net_header + (11))) =
      (r0);
  r0 = *(unsigned int *)((unsigned char *)l_native_header + (12));
  *((char *)((char *)l_net_header + (12))) =
      (((r0 << 4) & 0xf0L) |
    ((r0 >> 4) & 0xfL));
  *((char *)((char *)l_net_header + (13))) =
      ((r0 >> 8));
  *((char *)((char *)l_net_header + (14))) =
      ((r0 >> 24));
  *((char *)((char *)l_net_header + (15))) =
      ((r0 >> 16));
  r0 = *(unsigned int *)((unsigned char *)l_native_header + (16));
  *((char *)((char *)l_net_header + (16))) =
      ((r0 >> 8));
  *((char *)((char *)l_net_header + (17))) =
      (r0);
  *((char *)((char *)l_net_header + (18))) =
      ((r0 >> 24));
  *((char *)((char *)l_net_header + (19))) =
      ((r0 >> 16));
}


void
tcpHdrCopyUIn(l_native_header, l_net_header)  
  void* l_native_header;
  void* l_net_header;
{
  unsigned long r0;
  unsigned long r1;
  unsigned long r2;
  unsigned long r3;
  r0 = *(unsigned char *)((unsigned char *)l_net_header + (1));
  r1 = *(unsigned char *)((unsigned char *)l_net_header + (0));
  r2 = *(unsigned char *)((unsigned char *)l_net_header + (3));
  r3 = *(unsigned char *)((unsigned char *)l_net_header + (2));
  *((int *)((char *)l_native_header + (0))) =
      ((r0 & 0xffL) |
    ((r1 << 8) & 0xff00L) |
    ((r2 << 16) & 0xff0000L) |
    ((r3 << 24) & 0xff000000L));
  r0 = *(unsigned char *)((unsigned char *)l_net_header + (7));
  r1 = *(unsigned char *)((unsigned char *)l_net_header + (6));
  r2 = *(unsigned char *)((unsigned char *)l_net_header + (5));
  r3 = *(unsigned char *)((unsigned char *)l_net_header + (4));
  *((int *)((char *)l_native_header + (4))) =
      ((r0 & 0xffL) |
    ((r1 << 8) & 0xff00L) |
    ((r2 << 16) & 0xff0000L) |
    ((r3 << 24) & 0xff000000L));
  r0 = *(unsigned char *)((unsigned char *)l_net_header + (11));
  r1 = *(unsigned char *)((unsigned char *)l_net_header + (10));
  r2 = *(unsigned char *)((unsigned char *)l_net_header + (9));
  r3 = *(unsigned char *)((unsigned char *)l_net_header + (8));
  *((int *)((char *)l_native_header + (8))) =
      ((r0 & 0xffL) |
    ((r1 << 8) & 0xff00L) |
    ((r2 << 16) & 0xff0000L) |
    ((r3 << 24) & 0xff000000L));
  r0 = *(unsigned char *)((unsigned char *)l_net_header + (12));
  r1 = *(unsigned char *)((unsigned char *)l_net_header + (13));
  r2 = *(unsigned char *)((unsigned char *)l_net_header + (15));
  r3 = *(unsigned char *)((unsigned char *)l_net_header + (14));
  *((int *)((char *)l_native_header + (12))) =
      (((r0 << 4) & 0xf0L) |
    ((r0 >> 4) & 0xfL) |
    ((r1 << 8) & 0xff00L) |
    ((r2 << 16) & 0xff0000L) |
    ((r3 << 24) & 0xff000000L));
  r0 = *(unsigned char *)((unsigned char *)l_net_header + (17));
  r1 = *(unsigned char *)((unsigned char *)l_net_header + (16));
  r2 = *(unsigned char *)((unsigned char *)l_net_header + (19));
  r3 = *(unsigned char *)((unsigned char *)l_net_header + (18));
  *((int *)((char *)l_native_header + (16))) =
      ((r0 & 0xffL) |
    ((r1 << 8) & 0xff00L) |
    ((r2 << 16) & 0xff0000L) |
    ((r3 << 24) & 0xff000000L));
}


unsigned short
peekNetTcpCksum(l_net_header)  
  void* l_net_header;  
{
  unsigned long r0;
  unsigned long r1;
  unsigned short  r_peekNetTcpCksum;
  r0 = *(unsigned char *)((unsigned char *)l_net_header + (17));
  r1 = *(unsigned char *)((unsigned char *)l_net_header + (16));
  r_peekNetTcpCksum =((r0 & 0xffL) |
    ((r1 << 8) & 0xff00L));
  return r_peekNetTcpCksum;
}

#if defined(__STDC__) || defined(__GNUC__)
static void	dispPseudoHdr( IPpseudoHdr * );
static void	dispHdr( struct tcphdr * );
static bool	dumpBytes( char *, long, void * );
#else
static void	dispPseudoHdr( );
static void	dispHdr( );
static bool	dumpBytes( );
#endif


static void
dispPseudoHdr( h )
  IPpseudoHdr *h;
{
  printf("   IP pseudo header: src: %s  dst: %s  \n   z:  %d  p: %d len: %d\n",
	 ipHostStr(&h->src), ipHostStr(&h->dst), h->zero, h->prot,
	 ntohs(h->len));
}


static void
dispHdr( h )
  struct tcphdr *h;
{
    printf("sp: %d dp: %d ", h->th_sport, h->th_dport);
    printf("seq: %x ack: %x ", h->th_seq, h->th_ack);
    printf("off: %d w: %d ", h->th_off, h->th_win);
    printf("urp: %d fg: %s\n", h->th_urp, tcpFlagStr(h->th_flags));
}


static bool
dumpBytes( buf, len, foo )
  char *buf;
  long int len;
  VOID *foo;
{
    int i;

    for (i=0; i < len; i++) {
	printf("%x ",*(u_char *)buf++);
	if (i+1 % 50 == 0) {
	    putchar('\n');
	}
    }
    putchar('\n');
    return TRUE;
}

/*
 * tcpHdrStore -- 'arg' should point to an appropriate hdrStore_t 
 * note: '*hdr' and the length field of the pseudoheader in '*arg'
 * will be modified.
 */
void
tcpHdrStore(hdr, dst, len, arg)
    VOID *hdr;
    char *dst;
    long int len;
    VOID *arg;
{
    u_short 	sum;
    IPpseudoHdr	*pHdr;

    xAssert(len == sizeof(struct tcphdr));
    xIfTrace(tcpp, 6) {
	printf("Outgoing header\n");
	dispHdr((struct tcphdr *)hdr);
    }

    xTrace1(tcpp, 4, "Storing hdr with seq # %d",
	    ((struct tcphdr *)hdr)->th_seq);
    ((struct tcphdr *)hdr)->th_sum   = 0;      /* clear checksum field */

    if (( ((int)dst) % 4 ) == 0 )
      tcpHdrCopyOut( dst, (char*)hdr );
    else
      tcpHdrCopyUOut( dst, (char*)hdr );

    /*
     * Checksum
     */
#define msg (((hdrStore_t *)arg)->m)
    pHdr = ((hdrStore_t *)arg)->h;
    pHdr->len = htons(msgLen(msg));
    xIfTrace(tcpp, CKSUM_TRACE) {
	printf("Sent: ");
	dispPseudoHdr(pHdr);
    }
    sum = inCkSum(msg, (u_short *)pHdr, sizeof(IPpseudoHdr));

    xTrace1(tcpp, CKSUM_TRACE, "Cksum(x): %x", sum);

    pokeNetTcpCksum( (char *)dst, (char*)&sum );

    xIfTrace(tcpp, CKSUM_TRACE) {
	msgForEach(msg, dumpBytes, 0);
    }
    xAssert(inCkSum(msg, (u_short *)pHdr, sizeof(IPpseudoHdr)) == 0);
#undef msg
}


/*
 * tcpHdrLoad -- 'arg' should be a pointer to the message structure
 * The IP pseudoHdr should be attached as an attribute of the message.
 */
long
tcpHdrLoad(hdr, src, len, arg)
    VOID *hdr;
    char *src;
    long int len;
    VOID *arg;
{
    u_short	*pHdr;

    xAssert(len == sizeof(struct tcphdr));
    pHdr = (u_short *)msgGetAttr((Msg *)arg, 0);
    xAssert(pHdr);
    bcopy(src, hdr, len);
    xIfTrace(tcpp, CKSUM_TRACE) {
	printf("Received: ");
	dispPseudoHdr((IPpseudoHdr *)pHdr);
    }
    xTrace1(tcpp,CKSUM_TRACE,"Incoming cksum: %x",peekNetTcpCksum((char*)src));

#if 0
    {
	int i;
	u_char *c = src;

	for (i=0; i < len; i++) {
	    printf("%x ", *c++);
	}
	printf("\n");
    }
#endif
    if (( ((int)src) % 4 ) == 0 )
      tcpHdrCopyIn( (char*)hdr, (char *)src );
    else
      tcpHdrCopyUIn( (char*)hdr, (char *)src );

    xTrace1(tcpp, 4, "Loading hdr with seq # %d",
	    ((struct tcphdr *)hdr)->th_seq);

    ((struct tcphdr *)hdr)->th_sum = inCkSum((Msg *)arg, pHdr,
					      sizeof(IPpseudoHdr));

    xIfTrace(tcpp, 6) {
	printf("Incoming header\n");
	dispHdr((struct tcphdr *)hdr);
    }
#if 0
    {
	int i;
	u_char *c = hdr;

	for (i=0; i < len; i++) {
	    printf("%x ", *c++);
	}
	printf("\n");
    }
#endif
    return sizeof(struct tcphdr);
}


/*
 * tcpOptionsStore -- pads the options with zero bytes to a 4-byte boundary
 * 'arg' should point to an integer which will contain the actual number of
 * bytes in 'hdr'
 */
void
tcpOptionsStore( hdr, dst, len, arg )
  VOID *hdr;
  char *dst;
  long int len;
  VOID *arg;
{
    int hdrLen;	
    
    hdrLen = *(int *)arg;
    bcopy(hdr, dst, hdrLen);
    if (hdrLen % 4) {
	dst += hdrLen;
	do {
	    *dst++ = 0;
	    hdrLen++;
	} while (hdrLen % 4);
    }
}


long
tcpOptionsLoad( hdr, src, len, arg )
  VOID *hdr;
  char *src;
  long int len;
  VOID *arg;
{
    bcopy(src, hdr, len);
    return len;
}

