/*     
 * $RCSfile: ppp.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1994  Arizona Board of Regents
 *
 * $Log: ppp.c,v $
 * Revision 1.2  1994/10/27 17:27:28  andrey
 * removed non-ansi definitions
 *
 * Revision 1.1  1994/10/24  20:44:34  andrey
 * Initial revision
 *
 *
 * $Revision: 1.2 $
 * $Date: 1994/10/27 17:27:28 $
 */

#include "xkernel.h"
#include "ppp.h"
#include "lcp.h"
#include "arp.h"
#include "ppp_internal.h"
#include "ppp_constants.h"
#include "ppp_options.h"
#include "ppp_options.c"
#include "ppp_fsm.h"
#include "ppp_fsm.c"

int	tracepppp;
EscCharMap PPPDefaultCharMap = { 0x0000ff00 };

static void		getproc_protl(XObj);
static void		getproc_sessn(XObj);
extern xkern_return_t	pppProcessOpts(XObj);
static xkern_return_t   pppCloseProtl( XObj );
static xkern_return_t   pppCloseSessn( XObj );
static int              pppControlProtl( XObj, int, char *, int ) ;
static int              pppControlSessn( XObj, int, char *, int ) ;
static XObj             pppCreateSessn( XObj, XObj, XObj, ActiveID *, Part * );
static xkern_return_t   pppDemux( XObj, XObj, Msg * );
static long             pppHdrLoad( void *, char *, long, void * );
static void             pppHdrStore( void *, char *, long, void * );
static XObj             pppOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	pppOpenDone( XObj, XObj, XObj, XObj);
static xkern_return_t   pppOpenDisable( XObj, XObj, XObj, Part * );
static xkern_return_t   pppOpenEnable( XObj, XObj, XObj, Part * );
static xkern_return_t   pppOpenDisableAll( XObj, XObj );
static xkern_return_t   pppPop( XObj, XObj, Msg *, void * );
static xmsg_handle_t    pppPush( XObj, Msg * );

/*
 * hack, but it seems to fit the spec
 *  - add 0x8000 to a protocol to get its corresponding control protocol
 */
static ppp_protnum
lookup_control_protnum(ppp_protnum p)
{
  return p + 0x8000;
}

static int
is_control_prot(ppp_protnum pkey)
{
  if (
      ((pkey & 0x8000) == 0x8000) ||
      ((pkey & 0xC000) == 0xC000)
     )
    return 1;
  return 0;
}

static XObj
control_protocol( XObj self, ppp_protnum prot)
{
  PSTATE * ps = (PSTATE*)self->state;
  XObj C_Prot;

  if (mapResolve(ps->Control_Prot_map, &prot, &C_Prot) != XK_FAILURE)
    return C_Prot;
  return ERR_XOBJ;
}

static XObj
control_session( XObj self, ppp_protnum prot)
{
  PSTATE * ps = (PSTATE*)self->state;
  XObj C_Sess;

  if (mapResolve(ps->Control_Sess_map, &prot, &C_Sess) != XK_FAILURE)
    return C_Sess;
  return ERR_XOBJ;
}


/*
 * might want to check rules in ISO 3309 regarding
 * prot numbers, but we only know a few fixed values anyway
 */
static int
check_prot_num( long key, char * uppername, char * myname)
{
  if (key == -1) {
    xTrace2(pppp, TR_SOFT_ERROR, "PPP: prot num from %s to %s not found (-1)",
                uppername, myname);
    return 1;
  }
  if (key < 0 || key > 0xffff) {
    xTrace3(pppp, TR_SOFT_ERROR,
            "PPP: prot num (%d) from %s to %s is out of range",
            key, uppername, myname);
    return 1;
  }
  return 0;
}

static int
print_xobj( void * key, void * value, void * arg)
{
  ppp_protnum p = *(ppp_protnum*)key;
  XObj xo = (XObj) value;
  if (xIsSession(xo))
    xTrace3(pppp, TR_DETAILED, "%s (<- %s) : %x",
      xo->fullName, xGetUp(xo)->fullName, p);
  else
    xTrace2(pppp, TR_DETAILED, "%s proto : %x",
      xo->fullName, p);
  return MFE_CONTINUE;
}

static void
cache_control_prot( XObj self, ppp_protnum key, XObj pro)
{
  PSTATE * pstate = (PSTATE*)self->state;

  xAssert(xIsProtocol(pro));
  xTraceP2(self, TR_DETAILED, "cache_control_prot protocol %s prot %x",
          pro->fullName, key);

  switch(key) {
    case PPP_LCP:
    case PPP_PAP:
    case PPP_LQR:
    case PPP_CHAP:
    default:
      (void)mapBind(pstate->Control_Prot_map, &key, (int)pro);
  }
  xIfTraceP(self, TR_DETAILED) {
    printf("control protocol map is [");
    mapForEach(pstate->Control_Prot_map, print_xobj, 0);
    printf("]");
  }
}


static void
cache_control_sessn( XObj self, ppp_protnum key, XObj sess)
{
  PSTATE * pstate = (PSTATE*)self->state;

  xAssert(xIsSession(sess));
  xTraceP2(self, TR_DETAILED, "cache_control_sess %s prot %x",
          sess->fullName, key);

  switch(key) {
    case PPP_LCP:
    case PPP_PAP:
    case PPP_LQR:
    case PPP_CHAP:
    default:
      (void)mapBind(pstate->Control_Sess_map, &key, (int)sess);
  }
  xIfTraceP(self, TR_DETAILED) {
    printf("control session map is [");
    mapForEach(pstate->Control_Sess_map, print_xobj, 0);
    printf("]");
  }
}

#ifdef	not
static void
remove_control_prot( XObj self, ppp_protnum key)
{
  PSTATE * pstate = (PSTATE*)self->state;

  xTraceP1(self, TR_DETAILED, "remove_control_prot prot %x", key);
  (void)mapUnbind(pstate->Control_Prot_map, &key);
}

static void
remove_control_sess( XObj self, ppp_protnum key)
{
  PSTATE * pstate = (PSTATE*)self->state;

  xTraceP1(self, TR_DETAILED, "remove_control_sess prot %x", key);
  (void)mapUnbind(pstate->Control_Sess_map, &key);
}
#endif

/*
 * ppp_init
 */
void
ppp_init( XObj self)
{
  Part        part;
  PSTATE      *pstate;
  XObj        llp;

  xTraceP0(self, TR_GROSS_EVENTS, "init");
  xAssert(xIsProtocol(self));

  getproc_protl(self);
  pstate = X_NEW(PSTATE);
  self->state = (VOID *)pstate;

  PSTATE_init(pstate);

  pstate->phase = PPP_dead;

  (void)pppProcessOpts(self);

  if ( ! xIsProtocol(llp = xGetDown(self, 0)) ) {
    xError("PPP could not get lower protocol");
    return;
  }

  partInit(&part, 1);
  partPush(part, ANY_HOST, 0);

  if ( xOpenEnable(self, self, llp, &part) == XK_FAILURE ) {
    xTraceP0(self, TR_ALWAYS, "init: can't openenable lower protocol");
    xFree((char *) pstate);
    self->state = 0;
    return;
  }

  xTraceP0(self, TR_EVENTS, "init done");
}

/*
 * pppOpen
 */
static XObj
pppOpen( XObj self, XObj hlpRcv, XObj hlpType, Part *p)
{
  PSTATE  * pstate = (PSTATE *)self->state;
  IPhost  * remIPhost, * localIPhost;
  XObj    s, contp, conts;
  long	  caller_prot;
  ActiveID	activeid;

  xTraceP0(self, TR_MAJOR_EVENTS, "Open");

  caller_prot = relProtNum(hlpType, self);

  switch(caller_prot) {
    case PPP_IP:
    case PPP_OSI:
    case PPP_XEROXNSIDP:
    case PPP_DECNET_IV:
    case PPP_APPLETALK:
    case PPP_NOVEL_IPX:
    case PPP_VJC_TCPIP:
    case PPP_VJUC_TCPIP:
    case PPP_BPDU:
    case PPP_STR_PROT:
    case PPP_BANYAN_V:
    case PPP_HP802_1D:
    case PPP_LUXCOM:
    case PPP_SIGMA_NS:
      {
        ppp_protnum cp_num = lookup_control_protnum(caller_prot);
        if ((conts = control_session(self, cp_num)) == ERR_XOBJ) {
          if ((contp = control_protocol(self, cp_num)) == ERR_XOBJ) {
          xTraceP1(self, TR_ERRORS, "no control protocol for protocol %x",
                   caller_prot);
          return ERR_XOBJ;
          }
        }
      }
    break;

    /*
     * control protocols shouldn't be opening connections
     *   they just passively advertise their service
     *   via openEnable and are used by ppp
     */
    case PPP_IPCP:
    case PPP_ONLCP:
    case PPP_XNICP:
    case PPP_DPIVCP:
    case PPP_ACP:
    case PPP_NICP:
    case PPP_BNCP:
    case PPP_SPCP:
    case PPP_BVCP:

    case PPP_LCP:
    case PPP_PAP:
    case PPP_LQR:
    case PPP_CHAP:
      xTraceP1(self, TR_ERRORS, "control protocol %x doing open?",
              caller_prot);
      return ERR_XOBJ;

    default:
      xTraceP1(self, TR_ERRORS, "unknown protocol %x", caller_prot);

      return ERR_XOBJ;
  }

  if (pstate->LCP_session == ERR_XOBJ) {
    XObj ppp_lcp_sessn;
    XObj lcp_prot;
    void * ptr;
    ActiveID a;
    Part part;
    int len;

    a.prot = PPP_LCP;
    partInit(&part, 1);
    lcp_prot = control_protocol(self, a.prot);
    if (lcp_prot == ERR_XOBJ) {
      xTraceP1(self, TR_ERRORS, "no control protocol for %x", caller_prot);
      return ERR_XOBJ;
    }

    ppp_lcp_sessn = pppCreateSessn(self, lcp_prot, lcp_prot, &a, &part );
    xDuplicate(ppp_lcp_sessn);
    cache_control_sessn(self, PPP_LCP, ppp_lcp_sessn);

    if (xOpenDone(lcp_prot, ppp_lcp_sessn, self) == XK_FAILURE) {
      xTraceP0(self, TR_ERRORS, "can't open link control protocol");
      return ERR_XOBJ;
    }

    len = xControl(lcp_prot, PPP_NEGOTIATE, (void *)&ptr, sizeof(ptr));
    xTraceP1(self, TR_DETAILED, "negotiate returns %d", len);
    if (len == sizeof(ptr)) {
      xTraceP0(self, TR_DETAILED, "link control passed negotiation");
      pstate->LCP_session = ppp_lcp_sessn;
    }
    else {
      xTraceP0(self, TR_ERRORS, "link control failed negotiation");
      xClose(ppp_lcp_sessn);
      return ERR_XOBJ;
    }
  }
  /*
   * PPP session for protocol control-protocol needs to be created and the
   *   control protocol must be signaled to negotiate
   */
  if (conts == ERR_XOBJ) {
    XObj ppp_sessn;
    XObj cp_prot;
    void * ptr;
    ActiveID a;
    Part part;
    int len;

    partInit(&part, 1);
    a.prot = lookup_control_protnum(caller_prot);
    cp_prot = control_protocol(self, a.prot);
    if (cp_prot == ERR_XOBJ) {
      xTraceP1(self, TR_SOFT_ERROR, "no control protocol for %x", caller_prot);
      return ERR_XOBJ;
    }

    ppp_sessn = pppCreateSessn(self, cp_prot, cp_prot, &a, &part );
    xDuplicate(ppp_sessn);

    pstate->control_waiting = ppp_sessn;

    if (xOpenDone(cp_prot, ppp_sessn, self) == XK_FAILURE) {
      xTraceP1(self, TR_ERRORS, "can't open control protocol for %x",
              caller_prot);
      return ERR_XOBJ;
    }

    len = xControl(cp_prot, PPP_NEGOTIATE, (void *)&ptr, sizeof(ptr));
    xTraceP2(self, TR_DETAILED, "control prot %x negotiate returns %d",
            a.prot, len);
    if (len == sizeof(ptr)) {
      xTraceP0(self, TR_DETAILED, "control passed negotiation");
      cache_control_sessn(self, a.prot, pstate->control_waiting);
    }
    else {
      xTraceP0(self, TR_ERRORS, "control failed negotiation");
      xClose(ppp_sessn);
      return ERR_XOBJ;
    }
  } 

  /*
   * this address stuff must be bundled up
   * and modularized on a per-protocol basis
   * - way too much IP in here now
   */
  remIPhost = (IPhost *)partPop(p[0]);
  
  if (!IP_EQUAL(*remIPhost, pstate->IP_link.remote)) {
    xTraceP2(self, TR_ERRORS, "wanted remote %s, but I talk to remote %s",
        ipHostStr(remIPhost), ipHostStr(&pstate->IP_link.remote));
    return ERR_XOBJ;
  }
  if (partLen(p) > 1) {
    /*
     * Local participant has been explicitly specified
     */
    localIPhost = (IPhost *)partPop(p[1]);
    if (!IP_EQUAL(*localIPhost, pstate->IP_link.local)) {
      xTraceP2(self, TR_ERRORS, "wanted local %s, but I am %s",
        ipHostStr(localIPhost), ipHostStr(&pstate->IP_link.local));
      return ERR_XOBJ;
    }
  }
  xTraceP2(self, TR_DETAILED, "sends to %s, %d", ipHostStr(remIPhost), caller_prot);

  /*
   * map key is hlp prot number only
   */
  activeid.prot = caller_prot;
  s = pppCreateSessn(self, hlpRcv, hlpType, &activeid, p);
  xTraceP1(self, TR_DETAILED, "open returns %x", s);
  return s;
}

static xkern_return_t
pppOpenDone( XObj self, XObj lls, XObj llp, XObj hlpType)
{
  PSTATE * pstate = (PSTATE *) self->state;
  xTraceP0(self, TR_DETAILED, "OpenDone");
  pstate->Out_session = lls;
  return XK_SUCCESS;
}

static XObj
pppCreateSessn( XObj self, XObj hlpR, XObj hlpT, ActiveID * key, Part * p)
{
  XObj		s, lls;
  SSTATE	*sstate;
  PSTATE	*pstate;
  ETHhost	remote;

  xTraceP0(self, TR_EVENTS, "CreateSessn");
  pstate = (PSTATE *)self->state;

  if (mapResolve(pstate->Active_map, key, &s) != XK_FAILURE)
    return s;
  *((IPhost *) &remote) = pstate->IP_link.remote;
  remote.low = pstate->IP_link.port;
  partPush(p[0], &remote, sizeof(ETHhost));
  lls = xOpen(self, self, xGetDown(self, 0), p);
  if (lls == ERR_XOBJ) return ERR_XOBJ;

  s = xCreateSessn(getproc_sessn, hlpR, hlpT, self, 1, &lls);
  xAssert(xIsSession(s));

  s->binding = (Bind)mapBind(pstate->Active_map, (char *)key, (int)s);

  if ( s->binding == ERR_BIND ) {
      xTraceP0(self, TR_EVENTS, "CreateSessn - session already exists");
      pppCloseSessn(s);
      mapResolve(pstate->Active_map, key, &s);
      return s;
  }

  sstate = X_NEW(SSTATE);
  SSTATE_init(sstate);

  sstate->prot = key->prot;
  s->state = (char *)sstate;

  xTraceP0(self, TR_DETAILED, "Exiting pppCreateSessn");
  return s;
}

/*
 * pppControlSessn
 */
static int
pppControlSessn( XObj self, int opcode, char *buf, int len)
{
  xTraceP0(self, TR_EVENTS, "ControlSessn");

  switch(opcode) {

    /* just pass it on */
    case PPP_DONE_NEGOTIATION:
      xControl(self->myprotl, opcode, buf, len);
    break;

    case GETPARTICIPANTS:
    break;

    case GETMYHOST:
    break;

    case GETMYHOSTCOUNT:
      checkLen(len, sizeof(int));
      *(int *)buf = 1;
      return sizeof(int);
    break;

    case GETMAXPACKET:
      {
        int pksize;
        checkLen(len, sizeof(int));
        if (xControl(xGetDown(self, 0), GETMAXPACKET, (void *)&pksize,
                     sizeof(pksize)) != sizeof(pksize)) {
          *(int*)buf = PPP_MAXPACKET;
          return sizeof(int);
        }
        /* XXX header size is ppp_protnum - formalize this */
        *(int*)buf = pksize - sizeof(ppp_protnum) < PPP_MAXPACKET ?
                     pksize - sizeof(ppp_protnum) : PPP_MAXPACKET;
        return sizeof(int);
      }
    break;

    case GETOPTPACKET:
    break;


    case GETPEERHOST:
    case GETPEERHOSTCOUNT:
    case GETMYPROTO:
    case GETPEERPROTO:
    case FREERESOURCES:
    case SETNONBLOCKINGIO:
    default:
      xTraceP1(self, TR_MAJOR_EVENTS, "Unhandled opcode %d -- forwarding",
               opcode);
  }
  return xControl(xGetDown(self, 0), opcode, buf, len);
}

/*
 * pppControlProtl
 */
static int
pppControlProtl( XObj self, int opcode, char *buf, int len)
{
  PSTATE * ps = (PSTATE *)self->state;

  xTraceP0(self, TR_EVENTS, "ControlProtl");

  switch(opcode) {

    case PPP_DONE_NEGOTIATION:
      {
        ppp_protnum pn;
        SSTATE * sstate;
        if (ps->control_waiting == ERR_XOBJ) {
          xTraceP0(self, TR_ERRORS, "done nego. but no outstanding proto!");
          return 0;
        }
        sstate = (SSTATE *)ps->control_waiting->state;
        pn = sstate->prot;
        xTraceP1(self, TR_EVENTS, "control %x is done negotiation", pn);
        if (pn == PPP_LCP) {
          ps->LCP_session = ps->control_waiting;
        }
        cache_control_sessn(self, pn, ps->control_waiting);
        ps->control_waiting = ERR_XOBJ;
        return 1;
      }
    break;

    /* threee hacks to imitate arp */
    case RESOLVE:
      {
        ArpBinding * arpb = (ArpBinding *)buf;
        /* move ip address to mac address */
        *((IPhost *) &arpb->hw) = arpb->ip;
        arpb->hw.low = 0;
        return sizeof(ArpBinding);
      }
    break;
    case RRESOLVE:
      {
        ArpBinding * arpb = (ArpBinding *)buf;
        /* move mac address to ip address */
        arpb->ip = *((IPhost *) &arpb->hw);
        return sizeof(ArpBinding);
      }
    break;
    case ARP_GETMYBINDING:
      {
        ArpBinding * arpb = (ArpBinding *)buf;
        checkLen(len, sizeof(ArpBinding));
        *((IPhost *) &arpb->hw) = ps->IP_link.local;
        arpb->hw.low = 0;
        arpb->ip = ps->IP_link.local;
        return sizeof (ArpBinding);
      }
    break;

    case GETMYHOSTCOUNT:
      checkLen(len, sizeof(int));
      *(int *)buf = 1;
      return sizeof(int);
    break;

    case GETMAXPACKET:
      {
        int pksize;
        checkLen(len, sizeof(int));
        if (xControl(xGetDown(self, 0), GETMAXPACKET, (void *)&pksize,
                     sizeof(pksize)) != sizeof(pksize)) {
          *(int*)buf = PPP_MAXPACKET;
          return sizeof(int);
        }
        /* XXX header size is ppp_protnum - formalize this */
        *(int*)buf = pksize - sizeof(ppp_protnum) < PPP_MAXPACKET ? pksize - sizeof(ppp_protnum) : PPP_MAXPACKET;
        return sizeof(int);
      }
    break;

    case GETMYHOST:
    break;
      
    case GETOPTPACKET:
    break;

    default:
      xTraceP0(self, TR_MAJOR_EVENTS, "Unhandled opcode -- forwarding");
  }
  return xControl(xGetDown(self, 0), opcode, buf, len);
}
/*
 * pppOpenEnable
 */
static xkern_return_t
pppOpenEnable( XObj self, XObj hlpRcv, XObj hlpType, Part *p)
{
  ppp_protnum  key;
  PSTATE  *pstate;
  PassiveID pid;

  pstate = (PSTATE *)self->state;
  key = relProtNum(hlpType, self);

  xTraceP3(self, TR_EVENTS, "pppOpenEnable by (%s,%s) prot %x",
          hlpRcv->fullName, hlpType->fullName, key);

  if (check_prot_num(key, hlpType->name, self->name)) return XK_FAILURE;

  if (is_control_prot(key)) {
    cache_control_prot(self, key, hlpRcv);
  }
  pid.prot = key;
  return defaultOpenEnable(pstate->Passive_map, hlpRcv, hlpType, (VOID *)&pid);
}

/*
 * pppOpenDisable
 */
static xkern_return_t
pppOpenDisable( XObj self, XObj hlpRcv, XObj hlpType, Part *p)
{
  long  key;
  PSTATE  *pstate = (PSTATE *)self->state;

  xTraceP0(self, TR_EVENTS, "OpenDisable");
  pstate->phase = PPP_dead;

  key = relProtNum(hlpType, self);
  if (check_prot_num(key, hlpType->name, self->name)) return XK_FAILURE;
  return defaultOpenDisable(pstate->Passive_map, hlpRcv,hlpType, &key);
}

/*
 * pppOpenDisableAll
 */
static xkern_return_t
pppOpenDisableAll( XObj self, XObj hlpRcv)
{
  PSTATE  *pstate = (PSTATE *)self->state;

  xTraceP0(self, TR_EVENTS, "OpenDisableAll");
  pstate->phase = PPP_dead;

  return defaultOpenDisableAll(pstate->Passive_map, hlpRcv, 0);
}


/*
 * pppCloseSessn
 */
static xkern_return_t
pppCloseSessn(XObj self)
{
    xAssert(xIsSession(self));
    xTraceP1(self, TR_MAJOR_EVENTS, "PPP Closing session %x (in ppp)", self);
    xAssert(self->rcnt == 0);
    xFree((char *)self->state);
    return XK_SUCCESS;
}

/*
 * pppCloseProtl
 */
static xkern_return_t
pppCloseProtl( XObj self)
{
  PSTATE        *pstate;

  xAssert(xIsProtocol(self));
  xTraceP0(self, TR_MAJOR_EVENTS, "CloseProtl");
  if (((self->rcnt)) <= 0) {
    pstate = (PSTATE *) self->state;
    mapClose(pstate->Active_map);
    mapClose(pstate->Passive_map);
    xFree((char *) pstate);
  }
  return XK_SUCCESS;
}

static void
pppHdrStore( void * src, char * dst, long count, void * type)
{
  short n = htons(*(ppp_protnum *) src);

  xAssert(count == sizeof(ppp_protnum));
  bcopy((char *)&n, dst, count);
}

static long
pppHdrLoad( void * dst, char * src, long count, void * type)
{
  short n;

  xAssert(count == sizeof(ppp_protnum));
  bcopy(src, (char *)&n, count);
  *(ppp_protnum *)dst = ntohs(n);
  return sizeof(n);
}

/*
 * pppPush
 */
static xmsg_handle_t
pppPush( XObj self, Msg *msg)
{
  SSTATE      *sstate;
  XObj	      outsess;
  int	      d = msgLen(msg);


  xTraceP1(self, TR_EVENTS, "Push len %d", d);
  xAssert(xIsSession(self));
  outsess = ((PSTATE *)self->myprotl->state)->Out_session;
  sstate = (SSTATE *) self->state;

  msgPush(msg, pppHdrStore, &sstate->prot, sizeof(ppp_protnum), 0);

  if (outsess != ERR_XOBJ)
    return xPush(outsess, msg);
  else
    return xPush(xGetDown(self, 0), msg);
}

/*
 * pppDemux
 */
xkern_return_t
pppDemux( XObj   self, XObj   lls, Msg    *msg)
{
  XObj	session, Proto;
  PSTATE  *pstate = self->state;
  ppp_protnum  prot;
  Enable * e;
  ActiveID aid;
  PassiveID pid;
  int	      d = msgLen(msg);

  xTraceP1(self, TR_EVENTS, "Demux len %d", d);

  msgPop(msg, pppHdrLoad, &prot, sizeof(ppp_protnum), 0);
  if (check_prot_num(prot, lls->fullName, self->fullName)) {
    xTraceP1(self, TR_ERRORS, "bad protocol number: %x", prot);
    return XK_FAILURE;
  }
  aid.prot = prot;

  if (is_control_prot(prot)) {

    /* ensure that LCP is alive and well */
    if (prot != PPP_LCP) {
      if (pstate->LCP_session == ERR_XOBJ) {
        xTraceP1(self, TR_ERRORS, "protocol out of sequence (no lcp): %x",
                prot);
        return XK_FAILURE;
      }
    }

    if ((session = pstate->control_waiting) != ERR_XOBJ &&
        ((SSTATE*)(session->state))->prot == prot) {
    }
    else if ((Proto = control_protocol(self, prot)) != ERR_XOBJ) {
      if ((session = control_session(self, prot)) == ERR_XOBJ) {
        Part p[1];
        partInit(p, 1);

        session = pppCreateSessn(self, Proto, Proto, &aid, p);
        xDuplicate(session);
        pstate->control_waiting = session;
        if (session != ERR_XOBJ) {
          xOpenDone(Proto, session, self);
          xTraceP0(self, TR_EVENTS, "demux: sending message to new session");
        }
      }
    }
    else {
      xTraceP1(self, TR_ERRORS, "no OpenEnable for protocol: %x", prot);
      return XK_FAILURE;
    }
  }
  else {

    /*
     * check to see that associated control protocols are alive
     */
    ppp_protnum pcp = lookup_control_protnum(prot);

    if (pstate->LCP_session == ERR_XOBJ) {
      xTraceP1(self, TR_ERRORS, "protocol out of sequence (no LCP): %x",
              prot);
      return XK_FAILURE;
    }
    xIfTraceP(self, TR_DETAILED) {
      printf("control session map is [");
      mapForEach(pstate->Control_Sess_map, print_xobj, 0);
      printf("]");
    }

    if ((session = control_session(self, pcp)) == ERR_XOBJ) {
      xTraceP2(self, TR_ERRORS,
         "out of sequence mesg: no control session %x for protocol: %x",
         pcp, prot);
      return XK_FAILURE;
    }

    if (mapResolve(pstate->Active_map, &aid, &session) == XK_FAILURE) {
      pid.prot = prot;
      if (mapResolve(pstate->Passive_map, &pid, &e) == XK_SUCCESS) {
        Part p[1];
        partInit(p, 1);
  
        xTraceP1(self, TR_EVENTS, "openenable exists for prot: %x", prot);
        session = pppCreateSessn(self, e->hlpRcv, e->hlpType, &aid, p);
        if (session != ERR_XOBJ) {
          xOpenDone(e->hlpRcv, session, self);
          xTraceP0(self, TR_EVENTS, "demux: sending message to new session");
        }
      }
      else {
        xTraceP1(self, TR_ERRORS, "no OpenEnable for protocol: %x", prot);
        return XK_FAILURE;
      }
    }
  }
  xPop(session, self, msg, 0);
  return XK_SUCCESS;
}

/*
 * pppPop
 */
static xkern_return_t
pppPop( XObj self, XObj ds, Msg * msg, void *arg)
{
  xTraceP1(self, TR_EVENTS, "Pop, length = %d", msgLen(msg));
  xAssert(xIsSession(self));

  return xDemux(self, msg);
}

static void
getproc_protl( XObj s)
{
  xAssert(xIsProtocol(s));
  s->close = pppCloseProtl;
  s->control = pppControlProtl;
  s->open = pppOpen;
  s->opendone = pppOpenDone;
  s->openenable = pppOpenEnable;
  s->opendisable = pppOpenDisable;
  s->demux = pppDemux;
  s->opendisableall = pppOpenDisableAll;
}


static void
getproc_sessn( XObj s)
{
  xAssert(xIsSession(s));
  s->push = pppPush;
  s->pop = pppPop;
  s->control = pppControlSessn;
  s->close = pppCloseSessn;
}

