/*     
 * $RCSfile: divider.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Log: divider.c,v $
 * Revision 1.2  1994/05/06 20:27:17  menze
 *   [ 1994/05/05          menze ]
 *   Use msgShow rather than MSG_SHOW
 *
 * Revision 1.1  1994/04/18  22:02:04  slm
 * Initial revision
 *
 */

#include "xkernel.h"
#include "psync.h"
#include "common.h"
#include "divider.h"
#include "divider_internal.h"
#include "utils.c"

#define MSG_SHOW	msgShow

       void           divider_init(XObj);
static void           getproc_protl(XObj), getproc_sessn(XObj);		
static XObj           divider_open(XObj, XObj, XObj, Part *);
static xkern_return_t divider_openenable(XObj, XObj, XObj, Part *);
static xkern_return_t divider_demux(XObj, XObj, Msg *);
static int            divider_controlprotl(XObj, int, char *, int);
static xkern_return_t divider_pop(XObj, XObj, Msg *, VOID *);
static int            divider_controlsessn(XObj, int, char *, int);
static DEMUXvalue     *get_demuxvalue(XObj);
static long           rmsg_hdr_peek(VOID *, char *, long, VOID *);
static long           fmsg_hdr_peek(VOID *, char *, long, VOID *);
static void           call_demux(Event, VOID *);

int tracedividerp = 1;

void divider_init(self)
XObj self;
{
    DIVIDERstate *pstate;

    xAssert(xIsProtocol(self));
    xTrace0(dividerp, TR_GROSS_EVENTS, "DIVIDER init");
    if (!xIsProtocol(xGetDown(self,0)))
        Kabort("DIVIDER: could not get lower protocol PSYNC");

    getproc_protl(self);

    if ((pstate = (DIVIDERstate *)xMalloc(sizeof(DIVIDERstate))) == NULL)
	Kabort("DIVIDER: malloc failed in divider_init");
    self->state = (VOID *)pstate;
    pstate->map_openenable = mapCreate(100, sizeof(int));
    pstate->map_demux      = mapCreate(100, sizeof(Bindkey));
}

static void getproc_protl(self)
XObj self;
{
    xAssert(xIsProtocol(self));

    self->open       = divider_open;
    self->openenable = divider_openenable;
    self->demux      = divider_demux;
    self->control    = divider_controlprotl;
}

static void getproc_sessn(self)
XObj self;
{
    xAssert(xIsSession(self));

    self->pop     = divider_pop;
    self->control = divider_controlsessn;
}

static XObj divider_open(self, hlpRcv, hlpType, p)
XObj self, hlpRcv, hlpType;
Part *p;  /* p = [composer_id,port,psync1,...,psyncN,NULL] if hlp == COMPOSER */
          /* p = [psync,type1,...,typeN,NULL] if hlp != COMPOSER */
{
    DIVIDERstate *pstate = (DIVIDERstate *)self->state;
    Part         part[MAX_PART+1];
    XObj         sessn;
    int          *composer_id, *port, i, *type;
    PSYNCaddr    *curr_addr;
    Bindkey      bindkey;
    DEMUXvalue   *divider_sessn, *newval;

    xAssert(xIsProtocol(self));
    xTrace0(dividerp, TR_FUNCTIONAL_TRACE, "DIVIDER: entering divider_open");

    /*
     * Initialize the participant list size to MAX_PART+3 as is expected
     * by PSYNC.
     */
    partInit(part, MAX_PART+1);

    /*
     * if hlp = COMPOSER protocol
     *     open PSYNC with the part list as [psync0,psync1,...,psyncN]
     *     return the PSYNC Session
     * else
     *     create a session
     *     bind the msg types passed thru the partlist by hlp to the session
     *     return the DIVIDER session
     */
    if (hlpRcv == xGetProtlByName("composer")) {
        /*
         * From the partlist [composer_id,port,psync1,....,psyncN]
         * ignore the composer_id and port entries and send the rest.
         */
        composer_id = (int *)partPop(p[0]);
        port        = (int *)partPop(p[1]);
        xTrace2(dividerp, TR_DETAILED,
                "DIVIDER: composer_id = %d, port = %d", *composer_id, *port);

        for (i = 0; curr_addr = (PSYNCaddr *)partPop(p[i+2]); i++) {
	    xTrace3(dividerp, TR_DETAILED,
                    "DIVIDER: part %d, PSYNCport = %d, IPhost = %s",
                    i, curr_addr->port, ipHostStr(&curr_addr->host));
            partPush(part[i], (VOID *)curr_addr, sizeof(PSYNCaddr));
        }
	partPush(part[i], (VOID *)NULL, 0);

        /*
         * Open PSYNC session.  Return this session and not my own session to
	 * COMPOSER.
         */
        sessn = xOpen(self, self, xGetDown(self,0), part);
        pstate->psync_sessn = sessn;
        xTrace0(dividerp, TR_FUNCTIONAL_TRACE,
                "DIVIDER: exiting divider_open returning PSYNC sessn");
        /* Return the PSYNC session to COMPOSER.  *** UNCONVENTIONAL *** */
        return sessn;
    }
    else {
        /*
         * Some other protocol has done an open.  Partlist will be
         *     [psync,type1,...,typeN]
         */
	/*
	 * Create a DIVIDER session and set its down session to be the PSYNC
	 * session.
	 */
        sessn = xCreateSessn(getproc_sessn, hlpRcv, hlpType, self, 1,
			     &pstate->psync_sessn);

        /*
         * Make a key out of the PSYNC session and all the msg types, and bind
	 * it to the map in the form of linked list.
         */
        bindkey.system_id = *(int *)partPop(p[0]);
        xTrace1(dividerp, TR_DETAILED, "DIVIDER: system_id = %d",
		bindkey.system_id);
        for (i = 1; type = (int *)partPop(p[i]); i++) {
	    xTrace2(dividerp, TR_DETAILED, "DIVIDER: type %d = %s",
		    i, type2str(*type));
            bindkey.msg_type = *type;
            if ((newval = get_demuxvalue(sessn)) == NULL) {
		xError("DIVIDER: get_demuxvalue returned NULL in divider_open");
		return ERR_XOBJ;
	    }

            /*
             * Form a linked list of <system_id,msg_type> to the set of
	     * sessions to whom a message is to be popped.
             */
            /* `map_demux' is <system_id,msg_type> --> <divider_sessn> */
            if (mapResolve(pstate->map_demux,
                    (VOID *)&bindkey, (VOID **)&divider_sessn) == XK_FAILURE) {
                mapBind(pstate->map_demux, (VOID *)&bindkey, (VOID *)newval);
                newval->next = NULL;
            }
            else {	/* Add new value to list. */
                newval->next        = divider_sessn->next;
                divider_sessn->next = newval;
            }
        }
        xTrace0(dividerp, TR_FUNCTIONAL_TRACE,
                "DIVIDER: exiting divider_open returning DIVIDER sessn");

        /* Return the self session (i.e., a DIVIDER session) to the hlp. */
        return sessn;
    }
}

static xkern_return_t divider_openenable(self, hlpRcv, hlpType, p)
XObj self, hlpRcv, hlpType;
Part *p;	/* p = [composer_id,psync1] */
{
    Part      part[1];
    int       composer_id;
    PSYNCaddr *psyncaddr;

    xTrace0(dividerp, TR_FUNCTIONAL_TRACE,
	    "DIVIDER: entering divider_openenable");

    partInit(part, 1);
    composer_id = *(int *)partPop(p[0]);

    /* Bind the composer_id to the hlp. */
    /* `map_openenable' is <composer_id> --> <hlp> */
    mapBind(((DIVIDERstate *)self->state)->map_openenable, (VOID *)&composer_id,
	    (VOID *)hlpRcv);

    /* Openenable PSYNC by setting the part list as [psync1]. */
    psyncaddr = (PSYNCaddr *)partPop(p[1]);
    xTrace3(dividerp, TR_DETAILED,
	    "DIVIDER: composer_id = %d, PSYNCport = %d, IPhost = %s",
	    composer_id, psyncaddr->port, ipHostStr(&psyncaddr->host));
    partPush(part[0], (VOID *)psyncaddr, sizeof(PSYNCaddr));
    if (xOpenEnable(self, self, xGetDown(self,0), part) == XK_FAILURE) {
       xError("DIVIDER: error in openenable");
       return XK_FAILURE;
    }
    xTrace0(dividerp, TR_FUNCTIONAL_TRACE,
	    "DIVIDER: exiting divider_openenable");
    return XK_SUCCESS;
}

static xkern_return_t divider_demux(self, lls, msg)
XObj lls, self;
Msg *msg;
{
    DIVIDERstate *pstate = (DIVIDERstate *)self->state;
    CalDmx       arg;
    R_msg        rhdr;
    F_msg        fhdr;
    int          composer_id;
    XObj         hlp;
    Bindkey      bindkey;
    DEMUXvalue   *divider_sessn;
    Msg          tmpmsg;

    xTrace0(dividerp, TR_FUNCTIONAL_TRACE, "DIVIDER: entering divider_demux");

    xTrace3(dividerp, TR_FUNCTIONAL_TRACE,
            "DIVIDER: got msg %x of len = %ld, hdrlen = %d",
	    msg, msgLen(msg), sizeof(R_msg));
    /*
     * Peek into the header, r_msg_hdr = [node,first,type].  In this case
     * [first,type] form the TOTAL_ORDER header and [node] is a header pushed
     * by PSYNC.  The node field inserted by PSYNC is needed by TOTAL_ORDER.
     */
    msgPop(msg, rmsg_hdr_peek, (VOID *)&rhdr, sizeof(R_msg), (VOID *)NULL);
#ifdef XK_FIRST_INT
    if (rhdr.first == 1) {
#else
    if (rhdr.first == '1') {
#endif
        /*
	 * Sent by COMPOSER to set up connections (TOTAL_ORDER puts the first
	 * field as '0').  See the fmsg_hdr which is
	 * [node,first,port,composer_id].
         */
        xTrace0(dividerp, TR_EVENTS, "DIVIDER: received first msg");
        msgPop(msg, fmsg_hdr_peek, (VOID *)&fhdr, sizeof(F_msg), (VOID *)NULL);

        xTrace4(dividerp, TR_DETAILED,
       "DIVIDER: F_msg hdr: node = %d, first = %c, port = %d, composer_id = %d",
	        fhdr.node, fhdr.first, fhdr.port, fhdr.composer_id);

        /* Store the current port number into the state. */
        pstate->port = fhdr.port;
        /*
	 * Get the composer_id from the header and find the associated protocol.
         */
        composer_id = fhdr.composer_id;
        /* Has an openenable been done for the composer_id? */
        /* `map_openenable' is <composer_id> --> <hlp> */
        if (mapResolve(pstate->map_openenable, (VOID *)&composer_id,
		       (VOID **)&hlp) != XK_FAILURE) {
            /* Call the opendone of the hlp. */
            if (xOpenDone(hlp, lls, self) == XK_FAILURE) {
		xError("DIVIDER: error in demux");
		return XK_FAILURE;
	    }
	    /*
	     * Because PSYNC session (lls) is passed to opendone, its up is set
	     * to COMPOSER protocol (hlp); reset it to DIVIDER protocol (self).
	     */
	    xSetUp(lls, self);
        }
        if (xControl(lls, PSYNC_ACKMESSAGE, (char *)&fhdr.node,
		     sizeof(fhdr.node)) == -1) {
	    xError("DIVIDER: PSYNC_ACKMESSAGE returned error");
	    return XK_FAILURE;
        }
    }
    else {
        /*
         * Not the first message but a TOTAL_ORDER message.  Must have the
	 * type field.  Inspect it and pop it to the sessions needing it,
	 * determining, for this PSYNC session and type of message, what
	 * sessions have to be getting the message.
         */
        xTrace3(dividerp, TR_DETAILED,
                "DIVIDER: R_msg hdr: node = %d, first = %c, type = %s",
	        rhdr.node, rhdr.first, type2str(rhdr.type));
        bindkey.system_id = (int)lls;
        bindkey.msg_type  = rhdr.type;

        /* Find the DIVIDER session. */
        /* `map_demux' is <system_id,msg_type> --> <divider_sessn> */
        if (mapResolve(pstate->map_demux, (VOID *)&bindkey,
		       (VOID **)&divider_sessn) == XK_FAILURE) {
            xError("DIVIDER: can't find session to demux in divider_demux");
            return XK_FAILURE;
	}
	else {
	    xTrace1(dividerp, TR_DETAILED,
		    "DIVIDER: demuxing msg of len %d", msgLen(msg));
	    xIfTrace(dividerp, TR_DETAILED) {
	        MSG_SHOW(msg);
	    }
	    msgConstructCopy(&tmpmsg, msg);
	    xPop(divider_sessn->sessn, lls, msg, (VOID *)NULL);
            divider_sessn = divider_sessn->next;

            /* Give this message to all the sessions who need it. */
            while (divider_sessn) {
	        arg.sessn = divider_sessn->sessn;
		arg.lls   = lls;
                if ((arg.msg = (Msg *)xMalloc(sizeof(Msg))) == NULL) {
		    xError("DIVIDER: malloc failed in divider_demux");
		    return XK_FAILURE;
		}
	        msgConstructCopy(arg.msg, &tmpmsg);

	        /* send an ack ??? */
		/*
                 * Schedule a thread which will pass on the msg to the desired
		 * protocol.
                 */
	        evDetach(evSchedule(call_demux, (VOID *)&arg, 0));
                divider_sessn = divider_sessn->next;
            }
	    msgDestroy(&tmpmsg);
        }
    }
    xTrace0(dividerp, TR_FUNCTIONAL_TRACE, "DIVIDER: exiting divider_demux");
    return XK_SUCCESS;
}

static int divider_controlprotl(self, opcode, buf, len)
XObj self;
int  opcode, len;
char *buf;
{
    DIVIDERstate *pstate = (DIVIDERstate *)self->state;
    XObj         hlp, sessn;
    int          new_s, numtypes, i;
    Bindkey      bindkey;
    char         *typeptr;
    DEMUXvalue   *divider_sessn, *newval;

    xAssert(xIsProtocol(self));

    switch (opcode) {
        case DIVIDER_RESTORE_SESSN:	/* not used yet (not fixed) - slm */
            /* buf = {hlp,old_s,numleaves,leaves_mid} if hlp == COMPOSER */
	    /* buf = {hlp,new_s,numtypes,type1,...typeN} if hlp != COMPOSER */
            hlp = *(XObj *)buf;
            if (hlp == xGetProtlByName("composer")) {
                *(XObj *)buf = self;
                return xControl(xGetDown(self,0), PSYNC_RESTORESESSN, buf, len);
            }
            else {
                new_s = *(int *)(buf + sizeof(XObj));
                numtypes = *(int *)(buf + sizeof(XObj) + sizeof(int));
/*** ??? NULL POinter passed here but many not be the best thing ***/
		/*
		 * Create a DIVIDER session and set its down session to be
		 * NULL.
		 */
                sessn = xCreateSessn(getproc_sessn, hlp, hlp, self, 1,
				     (XObj *)NULL);
                bindkey.system_id = new_s;
                typeptr = (char *)(buf + sizeof(XObj) + 2 * sizeof(int));
                for (i = 0; i < numtypes; i++) {
                    bindkey.msg_type = *((int *)typeptr);
                    if ((newval = get_demuxvalue(sessn)) == NULL) {
		        xError(
	       "DIVIDER: get_demuxvalue returned NULL in divider_controlprotl");
		        return -1;
	            }
                   /* `map_demux' is <system_id,msg_type> --> <divider_sessn> */
                    if (mapResolve(pstate->map_demux, (VOID *)&bindkey,
				   (VOID **)&divider_sessn) == XK_FAILURE) {
/*** DOUBTFUL suddenly this map bind last param is worrying me ***/
                        mapBind(pstate->map_demux, (VOID *)&bindkey,
				(VOID *)newval);
                        newval->next = NULL;
                    }
                    else {
                        newval->next        = divider_sessn->next;
                        divider_sessn->next = newval;
                    }
                    typeptr += sizeof(int);
                }
                return (int)sessn;
            }
            return -1;

        case DIVIDER_GETCURRPORT:
	    /*
	     * Return port number for msg that just came in.  Required by
	     * COMPOSER while setting up connections for first msg.
	     */
	    checkLen(len, sizeof(pstate->port));
            *(int *)buf = pstate->port;
            return sizeof(pstate->port);

        default:
            return xControl(xGetDown(self,0), opcode, buf, len);
    }
}

/*
 * xkernel v3.2 requires that pop should not be bypassed.
 */
static xkern_return_t divider_pop(self, lls, msg, hdrp)
XObj self, lls;
Msg  *msg;
VOID *hdrp;
{
    xAssert(xIsSession(self));
    xTrace0(dividerp, TR_FUNCTIONAL_TRACE,
	    "DIVIDER: entering and exiting divider_pop");

    return xDemux(self, msg);
}

static int divider_controlsessn(self, opcode, buf, len)
XObj self;
int  opcode, len;
char *buf;
{
    xAssert(xIsSession(self));

    return xControl(xGetDown(self,0), opcode, buf, len);
}

static DEMUXvalue *get_demuxvalue(sessn)
XObj sessn;
{
    DEMUXvalue *new_sessn;

    if ((new_sessn = (DEMUXvalue *)xMalloc(sizeof(DEMUXvalue))) != NULL)
        new_sessn->sessn = sessn;
    return new_sessn;
}

/*
 * Peek into the header of a msg.
 */
static long rmsg_hdr_peek(hdr, src, len, arg)
VOID *hdr, *arg;
char *src;
long len;
{
    bcopy(src, (char *)hdr, len);
    return 0;
}

/*
 * Peek into the header of a msg.
 */
static long fmsg_hdr_peek(hdr, src, len, arg)
VOID *hdr, *arg;
char *src;
long len;
{
    bcopy(src, (char *)hdr, len);
    return 0;
}

/*
 * Scheduled to hand over copies of msg popped from PSYNC to the session above.
 */
static void call_demux(ev, arg)
Event ev;
VOID  *arg;
{
    CalDmx *caldmx = (CalDmx *)arg;

    xTrace0(dividerp, TR_DETAILED, "DIVIDER: call_demux: popping msg");
    xIfTrace(dividerp, TR_DETAILED) {
        MSG_SHOW(caldmx->msg);
    }
    xPop(caldmx->sessn, caldmx->lls, caldmx->msg, (VOID *)NULL);
    msgDestroy(caldmx->msg);
    xFree((char *)caldmx->msg);
}
