/*
 * rloginc.c
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993  Arizona Board of Regents
 *
 *
 */


#include "xkernel.h"
#include "vtap.h"
#include "rloginc_i.h"

#ifdef __STDC__

static void protlFuncInit(XObj);
static void sessnInit(XObj);
static XObj rlogincCreateSessn(XObj, XObj, XObj, XObj);
static XObj rlogincOpen(XObj, XObj, XObj, Part *);
static xkern_return_t rlogincOpenDisable(XObj, XObj, XObj, Part *);
static xkern_return_t rlogincOpenDone(XObj, XObj, XObj, XObj);
static xkern_return_t rlogincOpenEnable(XObj, XObj, XObj, Part *);
static int rlogincControlProtl(XObj, int, char *, int);
static xkern_return_t rlogincClose(XObj);
static xkern_return_t rlogincPop(XObj, XObj, Msg *, void *);
static xmsg_handle_t rlogincPush(XObj, Msg *);
static xkern_return_t rlogincProtlDemux(XObj, XObj, Msg *);
static xkern_return_t rlogincSessnDemux(XObj, XObj, Msg *);
static int rlogincControlSessn(XObj, int, char *, int);
void rloginc_init (XObj);

#else

static void protlFuncInit();
static void sessnInit();
static XObj rlogincCreateSessn();
static XObj rlogincOpen();
static xkern_return_t rlogincOpenDisable();
static xkern_return_t rlogincOpenDone();
static xkern_return_t rlogincOpenEnable();

#endif					/* __STDC__ */


int tracerlogincp;


void
rloginc_init(self)
    XObj self;
{
    PState *ps;

    xTrace0(rlogincp, TR_GROSS_EVENTS, "RLOGINC init");
    xAssert(xIsProtocol(self));
    if (!xIsProtocol(xGetDown(self, 0))) {
	xError("RLOGINC down vector is misconfigured (not a protocol)");
	return;
    }
    protlFuncInit(self);

    ps = X_NEW(PState);
    memset((char *)ps, 0, sizeof(PState));

    ps->activeMap = mapCreate(RLOGIN_ACT_MAP_SZ, sizeof(XObj));
    ps->passiveMap = mapCreate(RLOGIN_PAS_MAP_SZ, sizeof(XObj));
    self->state = (VOID *) ps;
}


static XObj
rlogincOpen(self, hlpRcv, hlpType, p)
    XObj self, hlpRcv, hlpType;
    Part *p;
{
    XObj rlogins, lls;
    PState *ps = (PState *) self->state;
    char *remuser, *locuser, *terminfo;
    Msg msg;
    static char str[2048], *pc;

    xTrace0(rlogincp, TR_MAJOR_EVENTS, "RLOGINC open");

    /* pop off the non-IP style rloginc participants */
    if (!(terminfo = (char *) partPop(p[1]))) {
        Kabort("RLOGINCC: couldn't not get participant[4] (terminfo)");
    }
    if (!(locuser = (char *) partPop(p[1]))) {
        Kabort("RLOGINCC: couldn't not get participant[3] (locuser)");
    }
    if (!(remuser = (char *) partPop(p[1]))) {
        Kabort("RLOGINCC: couldn't not get participant[2] (remuser)");
    }

    /* then do the open */
    if ((lls = xOpen(self, hlpType, xGetDown(self, 0), p)) == ERR_XOBJ) {
	xTrace0(rlogincp, TR_ERRORS, "rlogincOpen: could not open lower session");
	return ERR_XOBJ;
    }
    if (mapResolve(ps->activeMap, &lls, &rlogins) == XK_SUCCESS) {
	xTrace0(rlogincp, TR_MORE_EVENTS, "rlogincOpen -- found an existing one");
    } else {
	xTrace0(rlogincp, TR_MAJOR_EVENTS, "rlogincOpen -- creating a new one");
	if ((rlogins = rlogincCreateSessn(self, hlpRcv, hlpType, lls)) 
				== ERR_XOBJ) {
	    xClose(lls);
	}
    }

    /*
     * remote login server:
     *	remuser\0
     *	locuser\0
     *	terminal info\0
     *	data
     */

    /* build the rcmd string */
    pc = str;
    strcpy(pc, remuser);
    pc += strlen(remuser) + 1;

    strcpy(pc, locuser);
    pc += strlen(locuser) + 1;

    strcpy(pc, terminfo);
    pc += strlen(terminfo) + 1;

    msgConstructBuffer(&msg, str, pc - str);
    xPush(lls, &msg);
    msgDestroy(&msg);

    return rlogins;
}

static XObj
rlogincCreateSessn(self, hlpRcv, hlpType, lls)
    XObj self, hlpRcv, hlpType, lls;
{
    XObj s;
    PState *ps = (PState *) self->state;
    SState *ss;
    XTime t;

    if ((s = xCreateSessn(sessnInit, hlpRcv, hlpType, self, 1, &lls))
	== ERR_XOBJ) {
	xTrace0(rlogincp, TR_ERRORS, "create sessn fails in rlogincOpen");
	return ERR_XOBJ;
    }
    s->binding = mapBind(ps->activeMap, &lls, s);
    if (s->binding == ERR_BIND) {
	xTrace0(rlogincp, TR_ERRORS, "mapBind fails in rlogincCreateSessn");
	return ERR_XOBJ;
    }
    /*
     * The lower sessions' up field is made to point to this rloginc session
     * (not the protocol)
     */
    /*
    xSetUp(lls, s);
    */
    ss = X_NEW(SState);
    s->state = (VOID *) ss;
    xGetTime(&t);

    memset(ss, '\0', sizeof(SState));
    xAssert((ss->buf = malloc(SHELL_BUFFERSZ)) != NULL);

    xTrace1(rlogincp, TR_MAJOR_EVENTS, "RLOGINC open returns %x", s);
    return s;
}


static int
rlogincControlProtl(self, opcode, buf, len)
    XObj self;
    int opcode, len;
    char *buf;
{
    /*
     * All opcodes are forwarded to the lower protocol
     */
    return xControl(xGetDown(self, 0), opcode, buf, len);
}

static int
rlogincControlSessn(self, opcode, buf, len)
    XObj self;
    int opcode, len;
    char *buf;
{
    /*
     * All opcodes are forwarded to the lower protocol
     */
    return xControl(xGetDown(self, 0), opcode, buf, len);
}

static xkern_return_t
rlogincOpenEnable(self, hlpRcv, hlpType, p)
    XObj self, hlpRcv, hlpType;
    Part *p;
{
    PState *ps = (PState *) self->state;

    xTrace0(rlogincp, TR_MAJOR_EVENTS, "RLOGINC open enable");
    return defaultVirtualOpenEnable(self, ps->passiveMap, hlpRcv, hlpType,
				    self->down, p);
}


static xkern_return_t
rlogincOpenDisable(self, hlpRcv, hlpType, p)
    XObj self, hlpRcv, hlpType;
    Part *p;
{
    PState *ps = (PState *) self->state;

    return defaultVirtualOpenDisable(self, ps->passiveMap, hlpRcv, hlpType,
				     self->down, p);
}


static xkern_return_t
rlogincClose(s)
    XObj s;
{
    PState *ps;
    xkern_return_t res;
    XObj lls;

    xTrace1(rlogincp, TR_MAJOR_EVENTS, "RLOGINC close of session %x", s);
    xAssert(s->rcnt == 0);
    ps = (PState *) xMyProtl(s)->state;
    res = mapRemoveBinding(ps->activeMap, s->binding);
    xAssert(res != XK_FAILURE);
    lls = xGetDown(s, 0);
    xAssert(xIsSession(lls));
    xSetUp(lls, xMyProtl(s));
    xClose(lls);
    xDestroy(s);
    return XK_SUCCESS;
}

static xkern_return_t
rlogincOpenDone(self, lls, llp, hlpType)
    XObj self, lls, llp, hlpType;
{
    XObj s;
    PState *ps = (PState *) self->state;
    Enable *e;

    xTrace0(rlogincp, TR_MAJOR_EVENTS, "In RLOGINC openDone");
    if (self == hlpType) {
	xTrace0(rlogincp, TR_ERRORS, "self == hlpType in rlogincOpenDone");
	return XK_FAILURE;
    }
    /*
     * check for openEnable
     */

    if (mapResolve(ps->passiveMap, &hlpType, &e) == XK_FAILURE) {
	/*
	 * This shouldn't happen
	 */
	xTrace0(rlogincp, TR_ERRORS,
		"rlogincOpenDone: Couldn't find hlp for incoming session");
	return XK_FAILURE;
    }
    if ((s = rlogincCreateSessn(self, e->hlpRcv, e->hlpType, lls))
	== ERR_XOBJ) {
	return XK_FAILURE;
    }
    xDuplicate(lls);
    xTrace0(rlogincp, TR_EVENTS,
	    "rloginc Passively opened session successfully created");
    return xOpenDone(e->hlpRcv, s, self);
}


static xkern_return_t
rlogincProtlDemux(self, lls, m)
    XObj self, lls;
    Msg *m;
{
    PState *pstate = self->state;
    XObj prls;

    xTrace0(rlogincp, TR_MORE_EVENTS, "RLOGINC Entering Demux");

    /* we have to resolve first */
    if ( mapResolve(pstate->activeMap, &lls, &prls) == XK_FAILURE ) {
	xTrace0(rlogincp, TR_MAJOR_EVENTS, "RLOGINC Demux: bad mapResolve()");
	return XK_FAILURE;
    }

    xTrace0(rlogincp, TR_ERRORS, "rlogincProtlDemux called!!");
    return xPop(prls, lls, m, 0);
}


/*
 * rlogincPop and rlogincSessnDemux must be used (i.e., they can't be
 * bypassed) for the UPI reference count mechanism to work properly.
 */
static xkern_return_t
rlogincPop(s, lls, msg, h)
    XObj s, lls;
    Msg *msg;
    VOID *h;
{
    xTrace0(rlogincp, TR_EVENTS, "rloginc Pop");
    return xDemux(s, msg);
}

static xmsg_handle_t
rlogincPush( s, msg )
    XObj	s;
    Msg 	*msg;
{
  xTrace0(rlogincp, TR_EVENTS, "rlogincPush");
  return xPush(xGetDown(s, 0), msg);
}


static xkern_return_t
rlogincSessnDemux(self, lls, msg)
    XObj self, lls;
    Msg *msg;
{
    xTrace0(rlogincp, TR_EVENTS, "rloginc Session Demux");
    return xPop(self, lls, msg, 0);
}


static void
sessnInit(s)
    XObj s;
{
    xAssert(xIsSession(s));

    s->push = rlogincPush;
    s->pop = rlogincPop;
    s->close = rlogincClose;
    s->control = rlogincControlSessn;
    /*
     * RLOGINC sessions will look like a protocol to lower sessions, so we need
     * a demux function
     */
    /*
    s->demux = rlogincSessnDemux;
    */
}


static void
protlFuncInit(p)
    XObj p;
{
    xAssert(xIsProtocol(p));

    p->open = rlogincOpen;
    p->openenable = rlogincOpenEnable;
    p->opendisable = rlogincOpenDisable;
    p->demux = rlogincProtlDemux;
    p->opendone = rlogincOpenDone;
    p->control = rlogincControlProtl;
}
