/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.2
 * 
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: lance.c,v $
 *
 * HISTORY
 * $Log: lance.c,v $
 * Revision 1.27  1994/12/05 23:41:20  davidm
 * (bcopy_data_to_lance): Fixed bug that resulted in cases where 16 byte
 * chunks were written back to back (i.e., without the necessary 16 byte
 * gap inbetween).
 *
 * Revision 1.26  1994/12/05  01:52:28  davidm
 * Got rid of scout_swap_ipl() in favor of ScoutAtomicSequence().
 *
 * Revision 1.25  1994/10/26  19:10:03  davidm
 * Modified to take advantage of USC's ability to return values.
 *
 * (bcopy_data_to_lance): modified to reduce code size without reducing
 * performance significantly.
 *
 * Revision 1.24  1994/09/02  23:45:54  davidm
 * Minor changes.
 *
 * Revision 1.23  1994/08/27  01:52:56  davidm
 * Last version supporting the old msg-tool interface.  There are some
 * functions that have the "static" keyword commented out.  This is for
 * experimental purposes only.
 *
 * Revision 1.22  1994/08/10  17:23:10  davidm
 * (bzero_lance): simplified based on tracing results.
 *
 * Other minor fixes.
 *
 * Revision 1.21  1994/08/08  19:36:04  davidm
 * Improved prediction annotations based on profiling data (none of
 * the former predictions were wrong!).
 *
 * Revision 1.20  1994/06/23  15:20:55  davidm
 * Added predictions to if statements.
 *
 * Revision 1.19  1994/06/16  15:45:18  davidm
 * (lance_interrupt): renamed to interrupt() and made static; interrupt
 * handling cleaned up according to flamingo.c changes.
 *
 * Revision 1.18  1994/06/11  00:02:17  menze
 * Added copyright notice
 *
 * Revision 1.17  1994/05/24  23:48:43  druschel
 * Added handler argument to scout_device_set_interrupt().
 *
 * Revision 1.16  1994/05/15  00:18:25  davidm
 * Cleanup.
 *
 * Revision 1.15  1994/05/06  21:34:57  davidm
 * Last version that can choose easily between HSC and copy ops.
 *
 * Revision 1.14  1994/05/05  04:29:54  davidm
 * (attach): Interleaved buffers are now spaced apart as much as possible
 * to reduce the likelihood of trashing the cache.
 *
 * Revision 1.13  1994/05/02  20:43:21  davidm
 * Added support for HSC.
 *
 * Fixed nasty bzero_lance() bug (didn't work with interleaved frames...).
 *
 * Revision 1.12  1994/04/30  01:58:23  davidm
 * Interleaving of frames now controlled via XK_LANCE_INTERLEAVED.
 *
 * Revision 1.11  1994/04/25  21:47:30  davidm
 * (restart,init): init merged into restart.
 *
 * (tx_intr): fixed bug occuring when send queue was totally full.
 *
 * Revision 1.10  1994/04/24  05:28:22  davidm
 * (attach): ugly bug in tx buffer initialization fixed.
 *
 * Revision 1.9  1994/04/24  05:11:17  davidm
 * (lance_init): ps->my_ipl didn't get initialized.
 *
 * (write_frame): outgoing message got printed too early.
 *
 * Revision 1.8  1994/04/23  01:42:21  davidm
 * Added support for delayed copy-in on the receive path.
 *
 * Revision 1.7  1994/04/22  22:36:56  davidm
 * (attach): frame buffers are now interleaved so we don't waste half the
 * memory.
 *
 * Revision 1.6  1994/04/22  06:29:33  davidm
 * Modified to use xTraceP? where possible.
 *
 * Revision 1.5  1994/04/21  07:08:55  davidm
 * (rx_intr,push): Added code to play with LEDs where available.  Really
 * helps figuring out what's going on during a test compiled with
 * OPTIMIZE.  Extra overhead should be minimal.
 *
 * Revision 1.4  1994/04/21  03:44:55  davidm
 * Clean up.  Fixed bugs in send path (write_frame, write_frame_segment, and
 * bcopy_data_to_lance).
 *
 * Revision 1.3  1994/04/19  02:16:46  davidm
 * Minor changes.
 *
 * Revision 1.2  1994/04/15  20:46:57  davidm
 * First version that is half-way complete/working.
 *
 * Revision 1.1  1994/04/15  02:59:49  davidm
 * Initial revision
 *
 * History shortened.  Contributors so far: rcs, menze, af, rvb, mrt,
 * danner.
 */
/*
 * Driver for the AMD LANCE (Local Area Network Controller for
 * Ethernet).  It is based on the x-kernel version of the Mach LANCE
 * driver (xklance).
 */
/* 
 * Mach Operating System
 * Copyright (c) 1991,1990,1989 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */
#include "scout_device.h"
#include "scout_sync_arch.h"
#include "scout_thread.h"
#include "scout_util.h"
#include "xkernel.h"
#include "eth_i.h"
#include "lance.h"
#include "lance_mem.h"
#include "msg_internal.h"	/* XXX fixme! */

#define KB	1024

/*
 * Number of receive and transmit buffers:
 */
#ifdef XK_LANCE_INTERLEAVED
# define RLOG	6		/* 2**6 = 64 receive descriptors */
# define TLOG	4		/* 2**4 = 16 transmit descriptors */
#else
# define RLOG	5		/* 2**5 = 32 receive descriptors */
# define TLOG	3		/* 2**3 =  8 transmit descriptors */
#endif
#define NRX	(1<<RLOG) 	/* # of receive descriptors */
#define NTX	(1<<TLOG) 	/* # of transmit descriptors	*/

#define	LN_BUFFER_SIZE	1552	/* this permits 80 buffers (multiple of 16) */

/*
 * Memory addresses for LANCE are 24 bits wide.
 */
#define ADDR_LO(y)	((unsigned short)((u_long)(y) & 0xffff))
#define	ADDR_HI(y)	((unsigned short)(((u_long)(y)>>16) & 0xff))

/*
 * L2H(lo): converts a LANCE offset into the host memory address.
 * As every 32 bits contain only 16 bits of data, half the space
 * is wasted and the host sees the LANCE memory "diluted" by a factor
 * of two.
 */
#ifdef XK_DEBUG
# define L2H(lo) \
({ \
    xAssert(((lo)<<1) < 128*KB); \
    (void*)((char*) ps->ram + ((lo)<<1)); \
})
#else
# define L2H(lo) ((void*)((char*) ps->ram + ((lo)<<1)))
#endif
#define H2L(ha) (((long)(ha)-(long)ps->ram)/2)

typedef struct {
    bool		seated;			/* is interface seated? */
    Scout_IPL		my_ipl;			/* int. prio of LANCE */
    LANCE_Registers	*regs;
    u_long		*rom;			/* Ethernet address rom */
    int			ram_size;		/* RAM size as seen by LANCE */
    void		*ram;			/* LANCE shared memory */
    int			ram_sbrk;		/* first free RAM offset */
    int			init_block;		/* init block in shmem */
    int			rx_delayed_cnt;		/* # of pending copyins */
    int			rx_last;
    void		*rx_ring[NRX];
    void		*rx_buf [NRX];
    volatile int	tx_count;		/* # of packets enqueued */
    int			tx_last;		/* writer's index */
    int			tx_complete;		/* reader's index */
    void		*tx_ring[NTX];
    void		*tx_buf [NTX];
    Scout_FIFO		tx_queue;		/* queue of blocked senders */
    int			miss_cnt;
    int			restart_cnt;
    int			babbling_cnt;
    int			memerr_cnt;
    int			collision_cnt;
} PState;


int tracelancep = 0;

/*
 * These two functions are system specific:
 */
extern void lance_enable_dma (void *shared_memory_ptr);
extern const char *lance_config_str (void);

xkern_return_t	lance_init (XObj self);		/* give prototype */
static int	read_frame(XObj self, int index, int len);


typedef union {
    u_long  ul;
    void    *vp;
    u_long  *ulp;
    u_int   *uip;
    u_short *usp;
    u_char  *ucp;
} xAnyType;


static inline void*
bzero_lance(void *atp, int nbytes)
{
    int rest;
    xAnyType at;

    xTrace2(lancep, TR_DETAILED, "bzero_lance(at=%p,nbytes=%d)", at, nbytes);

    at.vp = atp;
    rest = 0x10 - (at.ul & 0xf);
    if (nbytes < rest) {
	rest = nbytes;
    } /* if */
    nbytes -= rest;
    if (XK_ALIGNED(at.ul, u_short)) {
	while (rest >= 2) {
	    rest -= 2; *at.usp++ = 0;
	} /* while */
    } /* if */
    while (rest) {
	--rest; *at.ucp++ = 0;
    } /* while */

    if (nbytes) {
	at.ul += 16;		/* skip gap */
	/*
	 * Usually, headers make up for a sizeable part of the
	 * 64 bytes minimum size imposed by Ethernet, so this
	 * case is very rare and we predict it false:
	 */
	if (nbytes >= 16 PREDICT_FALSE) {
	    xAssert((at.ul & 0xf) == 0);
	    while (nbytes >= 16) {
		nbytes -= 16;
		at.ulp[0] = at.ulp[1] = 0;
		at.ul += 32;
	    } /* while */
	} /* if */
	while (nbytes) {
	    --nbytes; *at.ucp++ = 0;
	} /* while */
    } /* if */
    return at.vp;
} /* bzero_lance */


/*
 * 4 32-bit words every 8 32-bit words.
 */
static inline void*
bcopy_data_to_lance(const void *fromp, void *top, int nbytes)
{
    xAnyType from, to;
    long rest, align, gap, rest_lt_16, nbytes_lt_rest;
#   define CP2L(a, unaligned) \
    { \
	u_long r0, r1; \
	from.ul -= (a); r0 = from.ulp[0]; r1 = from.ulp[1]; \
	while (nbytes >= 16) { \
	    from.ul += 16; nbytes -= 16; \
	    if (unaligned) { \
		to.ulp[0] = (r0>>8*(a)) | (r1<<(64-8*(a))); r0 = from.ulp[0]; \
		to.ulp[1] = (r1>>8*(a)) | (r0<<(64-8*(a))); r1 = from.ulp[1]; \
	    } else { \
		to.ulp[0] = r0; r0 = from.ulp[0]; \
		to.ulp[1] = r1; r1 = from.ulp[1]; \
	    } /* if */ \
	    to.ul += 32; \
	} /* while */ \
	from.ul += (a); \
    }

    xTrace3(lancep, TR_DETAILED,
	    "bcopy_data_to_lance(from=%p,to=%p,nbytes=%d)",
	    from, to, nbytes);

    to.vp = top;
    gap   = 0x10;
    align = (to.ul & 0xf);
    from.vp = (void*) fromp;
    rest = gap - align;			/* rest in range 1..15 */

    rest_lt_16 = rest < gap;
    nbytes_lt_rest = nbytes < rest;

    if (rest_lt_16) {
	if (nbytes_lt_rest) {
	    rest = nbytes;
	    gap  = 0;
	} /* if */

	bcopy(from.vp, to.vp, rest);
	nbytes  -= rest;
	from.ul += rest;
	to.ul   += rest + gap;
    } /* if */

    align = from.ul & 0x7;

    if (nbytes >= 16) {
	xAssert((to.ul & 0xf) == 0);
	if (align) {
	    CP2L(align, 1);
	} else {
	    CP2L(0, 0);
	} /* if */
    } /* if */

    if (nbytes) {
	bcopy(from.vp, to.vp, nbytes);
	to.ul += nbytes;
    } /* if */

    return to.vp;
} /* bcopy_data_to_lance */


static inline void
bcopy_data_to_host(const void *fromp, void *top, int nbytes)
{
    xAnyType from, to;
    u_long r0, r1;

    from.vp = (void*)fromp; to.vp = top;

    xTrace3(lancep, TR_DETAILED, "bcopy_data_to_host(from=%p,to=%p,nbytes=%d)",
	    from, to, nbytes);
    xAssert((from.ul & 0xf) == 0 && XK_ALIGNED(to.ul, u_long));

    r0 = from.ulp[0]; r1 = from.ulp[1];
    while (nbytes >= 16) {
	from.ul += 32; nbytes -= 16;
	to.ulp[0] = r0; r0 = from.ulp[0];
	to.ulp[1] = r1; r1 = from.ulp[1];
	to.ul += 16;
    } /* while */
    while (nbytes) {
	--nbytes; *to.ucp++ = *from.ucp++;
    } /* while */
} /* bcopy_data_to_host */


static void
bcopy_desc_to_host(const void *from, void *to, int nbytes)
{
    const u_int *sp;
    u_short *dp;
    int nshorts;

    xAssert((nbytes & 1) == 0);
    
    xTrace3(lancep, TR_DETAILED, "bcopy_desc_to_host(from=%p,to=%p,nbytes=%d)",
	    from, to, nbytes);

    nshorts = nbytes >> 1;
    sp = from;
    dp = to;
    while (nshorts--) {
	*dp++ = (u_short) *sp++;
    } /* while */
} /* bcopy_desc_to_host */


static void
bcopy_desc_to_lance(const void *from, void *to, int nbytes)
{
    u_short *sp = (u_short*) from;
    u_int   *dp = (u_int*) to;
    int nshorts;

    xAssert((nbytes & 1) == 0);
    
    xTrace3(lancep, TR_DETAILED,
	    "bcopy_desc_to_lance(from=%p,to=%p,nbytes=%d)", from, to, nbytes);

    nshorts = nbytes >> 1;
    while (nshorts--) {
	*dp++ = (u_int) *sp++;
    } /* while */
} /* bcopy_desc_to_lance */


static int
shmalloc(PState *ps, int size)
{
    int off;

    off = ps->ram_sbrk;
    ps->ram_sbrk += size;
    if (ps->ram_sbrk >= ps->ram_size) {
	scout_panic("ln: shmalloc: out of shared memory!");
    } /* if */
    return off;
} /* shmalloc */


static void
write_reg(LANCE_Registers *reg, int csr, u_short val)
{
    static int selected_csr = -1;
    int i = 0;

    if (csr != selected_csr) {
	while ((reg->rap & 0xffff) != csr) {
	    reg->rap = csr;
	    scout_flush_write_to_memory();
	    if (++i > 10000) {
		xTrace2(lancep, TR_ERRORS,
			"ln: unable to change RAP from %#x to %#x.",
			reg->rap, csr);
		xError("ln: RAP did not settle");
		return;
	    } /* if */
	    scout_udelay(100);
	} /* while */
	selected_csr = csr;
	i = 0;
    } /* if */
    while ((reg->rdp & 0xffff) != val) {
	reg->rdp = val;
	scout_flush_write_to_memory();
	if (++i > 10000) {
	    xTrace2(lancep, TR_ERRORS,
		    "ln: unable to change RDP from %#x to %#x.",
		    reg->rap, val);
	    xError("ln: RDP did not settle");
	    return;
	} /* if */
	scout_udelay(100);
    } /* while */
} /* write_reg */


/*
 * Initialize a ring descriptor and allocate a receive/transmit
 * buffer for it.  Finally, the descriptor is copied into
 * LANCE shared memory.
 */
static void
mkdesc(void *shm_desc, int frame)
{
    LANCE_Descriptor desc;

    /*
     * Build a descriptor pointing to it:
     */
    desc.addr_low = ADDR_LO(frame);
    desc.addr_hi  = ADDR_HI(frame);
    desc.status   = 0;
    desc.buffer_size = -LN_BUFFER_SIZE;	/* LANCE uses an up counter... */
    desc.u.bits = 0;

    /*
     * Copy the descriptor to LANCE shared memory:
     */
    bcopy_desc_to_lance(&desc, shm_desc, sizeof(desc));
    scout_flush_write_to_memory();
} /* mkdesc */


static bool
ship_init_block(XObj self)
{
    PState *ps = (PState*) self->state;
    LANCE_Registers *regs = ps->regs;
    int i;

    xTraceP1(self, TR_FUNCTIONAL_TRACE, "ship_init_block(block=%d)",
	     ps->init_block);

    /*
     * Tell LANCE where init block is located at:
     */
    write_reg(regs, CSR1, ADDR_LO(ps->init_block));
    write_reg(regs, CSR2, ADDR_HI(ps->init_block));

    /*
     * Start INIT sequence now and make sure CSR0 is selected.  From
     * now on, we will access CSR0 only and will do so directly
     * through regs->rdp.
     */
    regs->rap = CSR0;
    scout_flush_write_to_memory();

    regs->rdp = (LN_CSR0_IDON | LN_CSR0_INIT);	/* clear IDON & start init */
    scout_flush_write_to_memory();

    /* give it plenty of time to settle */
    for (i = 0; i < 10000; ++i) {
	scout_udelay(100);
	if (regs->rdp & LN_CSR0_IDON) {
	    break;
	} /* if */
    } /* for */

    /* make sure got out okay: */
    if (!(regs->rdp & LN_CSR0_IDON)) {
	if (regs->rdp & LN_CSR0_ERR) {
	    xTraceP1(self, TR_ERRORS, "initialization error, csr=%#04x",
		     regs->rdp & 0xffff);
	} /* if */
	xError("ln: attach: cannot initialize");
	return FALSE;
    } /* if */
    return TRUE;
} /* ship_init_block */


/*
 * Restart LANCE.
 *
 * Used on those errors that hang the chip.
 *
 * After stopping the chip and effectively turning off the LANCE, we
 * release all pending buffers and cause the chip to init itself.  We
 * do not enable interrupts here.
 */
static void
restart(XObj self)
{
    PState *ps = (PState*) self->state;
    LANCE_Registers *regs = ps->regs;
    Scout_Thread *t;
    int i;

    /* stop the chip: */
    write_reg(regs, CSR0, LN_CSR0_STOP);

    ++ps->restart_cnt;

    xTraceP1(self, TR_ERRORS, "%d restarts", ps->restart_cnt);

    /* free up any outgoing messages currently in queue: */
    while (ps->tx_count) {
	--ps->tx_count;
	t = scout_fifo_remove(&ps->tx_queue);
	if (t) {
	    xTraceP1(self, TR_FUNCTIONAL_TRACE, "restart: waking thread %s",
		     scout_thread_get_name(scout_thread_self()));
	    scout_thread_async_wakeup(t);
	} /*  if */
    } /* while */

    /* initialize the chip again, no need to reload init block address: */
    ship_init_block(self);

    /*
     * Mark all rx buffers to be owned by LANCE, tx buffers owned
     * by host.
     */
    for (i = 0; i < NRX; ++i) {
	set_desc_status(ps->rx_ring[i], LN_RSTATE_OWN);
    } /* for */
    for (i = 0; i < NTX; ++i) {
	set_desc_status(ps->tx_ring[i], 0);
    } /* for */
    scout_flush_write_to_memory();

    ps->tx_complete = ps->tx_last = ps->rx_last = 0;

    /* start chip & enable interrupts: */
    regs->rdp = LN_CSR0_STRT | LN_CSR0_INEA | LN_CSR0_TDMD;
    scout_flush_write_to_memory();
} /* restart */


static bool
attach(XObj self)
{
    PState *ps = (PState*) self->state;
    LANCE_Registers *regs = ps->regs;
    LANCE_Init_Block init_block;
    void *desc;
    int i, lp, fp;

    /*
     * Get a 128 KB buffer that is aligned to 128 KB (bottom 16 bits == 0).
     *
     * CAVEAT: As the host memory is dilated by a factor of 2, this
     * looks to LANCE as if it had 64 KB only.  It is possible to
     * work around this by interleaving frames (option XK_LANCE_INTERLEAVED).
     */
    ps->ram = scout_malloc_aligned(128*KB, 17);
    bzero(ps->ram, 128*KB);
    ps->ram_size = 64*KB;
    ps->ram_sbrk = 0;
    scout_fifo_init(&ps->tx_queue);

    /* tell the system where we allocated the shared memory at: */
    lance_enable_dma(ps->ram);

    /* reset the interface, and make sure we really do it! */
    write_reg(regs, CSR0, LN_CSR0_STOP);

    printf("ln: %s Ethernet at [%02x:%02x:%02x:%02x:%02x:%02x]\n",
	   lance_config_str(),
	   (u_int) ps->rom[0] & 0xff, (u_int) ps->rom[1] & 0xff,
	   (u_int) ps->rom[2] & 0xff, (u_int) ps->rom[3] & 0xff,
	   (u_int) ps->rom[4] & 0xff, (u_int) ps->rom[5] & 0xff);

    /* allocate init block in shared memory: */
    ps->init_block = shmalloc(ps, sizeof(LANCE_Init_Block));

    /*
     * Setup init block on stack, then copy everything into the
     * LANCE shared memory:
     */
    bzero(&init_block, sizeof(init_block));
    init_block.mode = 0;
    init_block.phys_addr_high = (ps->rom[0] & 0xff) | (ps->rom[1] & 0xff) << 8;
    init_block.phys_addr_med  = (ps->rom[2] & 0xff) | (ps->rom[3] & 0xff) << 8;
    init_block.phys_addr_low  = (ps->rom[4] & 0xff) | (ps->rom[5] & 0xff) << 8;

    /*
     * Allocate both descriptor rings at once.  Note that the quadword
     * alignment requirement is inherent in the way we perform
     * allocation, but it does depend on the size of the init block.
     */
    lp = shmalloc(ps, sizeof(LANCE_Descriptor) * (NTX + NRX));

    /*
     * Initialize the buffer descriptors
     */
    init_block.recv_ring_pointer_lo = ADDR_LO(lp);
    init_block.recv_ring_pointer_hi = ADDR_HI(lp);
    init_block.recv_ring_len = RLOG;

#ifdef XK_LANCE_INTERLEAVED
    for (i = 0; i < NRX/2; ++i) {
	desc = L2H(lp);
	/* allocate enough for two interleaved buffers: */
	fp = shmalloc(ps, LN_BUFFER_SIZE + 16);
	xAssert(((long) desc & 0xf) == 0 && ((long) L2H(fp) & 0xf) == 0);
	mkdesc(desc, fp);
	ps->rx_ring[i] = desc;
	ps->rx_buf[i]  = L2H(fp);
	lp += sizeof(LANCE_Descriptor);
    } /* for */
    for (i = NRX/2; i < NRX; ++i) {
	desc = L2H(lp);
	/* allocate enough for two interleaved buffers: */
	fp = H2L(ps->rx_buf[i - NRX/2]) + 8;
	xAssert(((long) desc & 0xf) == 0 && ((long) L2H(fp) & 0xf) == 0);
	mkdesc(desc, fp);
	ps->rx_ring[i] = desc;
	ps->rx_buf[i]  = L2H(fp);
	lp += sizeof(LANCE_Descriptor);
    } /* for */
#else
    for (i = 0; i < NRX; ++i) {
	desc = L2H(lp);
	fp = shmalloc(ps, LN_BUFFER_SIZE);
	xAssert(((long) desc & 0xf) == 0 && ((long) L2H(fp) & 0xf) == 0);
	mkdesc(desc, fp);
	ps->rx_ring[i] = desc;
	ps->rx_buf[i]  = L2H(fp);
	lp += sizeof(LANCE_Descriptor);
    } /* for */
#endif /* XK_LANCE_INTERLEAVED */

    init_block.xmit_ring_pointer_lo = ADDR_LO(lp);
    init_block.xmit_ring_pointer_hi = ADDR_HI(lp);
    init_block.xmit_ring_len = TLOG;

#ifdef XK_LANCE_INTERLEAVED
    for (i = 0; i < NTX/2; ++i) {
	desc = L2H(lp);
	/* allocate enough for two interleaved buffers: */
	fp = shmalloc(ps, LN_BUFFER_SIZE + 16);
	xAssert(((long) desc & 0xf) == 0 && ((long) L2H(fp) & 0xf) == 0);
	mkdesc(desc, fp);
	ps->tx_ring[i] = desc;
	ps->tx_buf[i]  = L2H(fp);
	lp += sizeof(LANCE_Descriptor);
    } /* for */
    for (i = NTX/2; i < NTX; ++i) {
	desc = L2H(lp);
	/* allocate enough for two interleaved buffers: */
	fp = H2L(ps->tx_buf[i - NTX/2]) + 8;
	xAssert(((long) desc & 0xf) == 0 && ((long) L2H(fp) & 0xf) == 0);
	mkdesc(desc, fp);
	ps->tx_ring[i] = desc;
	ps->tx_buf[i]  = L2H(fp);
	lp += sizeof(LANCE_Descriptor);
    } /* for */
#else
    for (i = 0; i < NTX; ++i) {
	desc = L2H(lp);
	fp = shmalloc(ps, LN_BUFFER_SIZE);
	xAssert(((long) desc & 0xf) == 0 && ((long) L2H(fp) & 0xf) == 0);
	mkdesc(desc, fp);
	ps->tx_ring[i] = desc;
	ps->tx_buf[i]  = L2H(fp);
	lp += sizeof(LANCE_Descriptor);
    } /* for */
#endif /* XK_LANCE_INTERLEAVED */
    /*
     * No logical address filtering
     */
    init_block.logical_addr_filter0 = 0;
    init_block.logical_addr_filter1 = 0;
    init_block.logical_addr_filter2 = 0;
    init_block.logical_addr_filter3 = 0;

    /*
     * Move init block into lance space
     */
    bcopy_desc_to_lance(&init_block, L2H(ps->init_block), sizeof(init_block));
    scout_flush_write_to_memory();

    restart(self);
    return TRUE;
} /* attach */


/*
 * Move an x-kernel message into LANCE space
 */
static inline int
write_frame(Msg *msg, void *sbuf, ETHhdr *hdr)
{
    MsgIterator mi;
    u_char *dst = sbuf;
    void *src;
    long len;
	
    /* write ETH header first: */
    src = hdr;
    len = sizeof(ETHhdr);

    msgIterateInit(&mi, msg);
    do {
	dst = bcopy_data_to_lance(src, dst, len);
    } while ((src = msgIterateNext(&mi, &len)) != 0);
    msgIterateDone(&mi);

    xIfTrace(lancep, TR_FULL_TRACE) {
	u_char *mp;

	printf("ln: write_frame: outgoing message:\n");
	msgShow(msg);

	mp = sbuf;
	printf("ln: write_frame: dst=[%02x:%02x:%02x:%02x:%02x:%02x], "
	       "src=[%02x:%02x:%02x:%02x:%02x:%02x], type=[%02x:%02x]\n",
	       "ln: write_frame: data=[%02x:%02x:%02x:%02x:%02x:%02x:"
	       "%02x:%02x..]\n",
	       mp[0], mp[1], mp[2], mp[3], mp[4], mp[5],
	       mp[6], mp[7], mp[8], mp[9], mp[10],mp[11], mp[12], mp[13],
	       mp[14], mp[15], mp[16], mp[17], mp[18], mp[19], mp[20], mp[21]);
    } /* if */

    len = msgLen(msg) + sizeof(ETHhdr);
    if (len < LN_MINBUF_NOCH) {
	/*
	 * LANCE needs at least this much data in a packet.  Zero out
	 * to that size to avoid security leak:
	 */
	bzero_lance(dst, LN_MINBUF_NOCH - len);
	len = LN_MINBUF_NOCH;
    } /* if */
    return len;
} /* write_frame */


/*static*/ xmsg_handle_t
lance_push(XObj s, Msg *msg)
{
    XObj self = xMyProtl(s);
    PState *ps = (PState*) self->state;
    ETHhdr *hdr = (ETHhdr*) msgGetAttr(msg, 0);
    int index, len, restart_cnt;

    if (ps->tx_count >= NTX PREDICT_FALSE) {
	do {
	    /* wait for a tx slot to become available: */
	    xTraceP2(self, TR_EVENTS,
		     "push: %s waiting for space (tx_count=%d)",
		     scout_thread_get_name(scout_thread_self), ps->tx_count);
	    scout_fifo_append(&ps->tx_queue, scout_thread_self());
	    scout_thread_suspend();
	} while (ps->tx_count >= NTX);
    } /* if */

    scout_set_led(SCOUT_LED_LAN_TX);

    xTraceP2(self, TR_EVENTS, "push: %s got send buffer (tx_count=%d)",
	     scout_thread_get_name(scout_thread_self), ps->tx_count);

    restart_cnt = ps->restart_cnt;
    index = ps->tx_last;
    xAssert(ps->tx_buf[index] >= ps->ram
	    && ps->tx_buf[index] < ps->ram + 2 * ps->ram_size);
    len = write_frame(msg, ps->tx_buf[index], hdr);

    /*
     * Hand over buffer to LANCE.  Must copyin/out, set len, and---in
     * a separate step---set the OWN flag.  We do not do chaining.
     */
    set_desc_buffer_size(ps->tx_ring[index], -len);
    set_desc_status(ps->tx_ring[index], LN_TSTATE_STP | LN_TSTATE_ENP);
    scout_flush_write_to_memory();

    set_desc_status(ps->tx_ring[index],
		    LN_TSTATE_STP | LN_TSTATE_ENP | LN_TSTATE_OWN);
    scout_flush_write_to_memory();

    /* tell chip to rescan descriptor buffers: */
    ps->regs->rdp = LN_CSR0_TDMD | LN_CSR0_INEA;
    scout_flush_write_to_memory();

    /* if this is a broadcast or multicast packet, loop it back: */
    if (ETH_ADS_MCAST(hdr->dst) PREDICT_FALSE) {
	read_frame(self, -index, len);
    } /* if */

    /*
     * If there hasn't been a restart, update tx_last (otherwise we
     * just silently drop this frame):
     */
    ScoutAtomicSequence(if (restart_cnt == ps->restart_cnt) {
			    ps->tx_last = (index + 1) % NTX;
			} /* if */);

    /*
     * It is important to do this last.  Otherwise, tx_intr() might
     * free the frame before we had a chance to copy it out again
     * with read_frame() in the case of a multicast packet.
     */
    ScoutAtomicSequence(++ps->tx_count);

    return XMSG_NULL_HANDLE;
} /* push */


/*static*/ void
copyin(XShepherdBuffer *buf, void *frame, int len)
{
    bcopy_data_to_host(frame, buf->data, len);
    msgPopDiscard(&buf->upmsg, sizeof(ETHhdr));
    msgSetAttr(&buf->upmsg, 0, buf->data, sizeof(ETHhdr));
} /* copyin */


#ifdef XK_SHEPHERD_PROLOG

/*static*/ void
delayed_copyin(XShepherdBuffer *buf, void *arg)
{
    XObj self = buf->driverProtl;
    PState *ps = (PState*) self->state;
    int index = (long) arg;

    xTraceP2(self, TR_FULL_TRACE, "delayed_copyin(i=%d,len=%d)",
	     index, msgLen(&buf->upmsg));

    copyin(buf, ps->rx_buf[index], msgLen(&buf->upmsg));

    /* return descriptor to LANCE: */
    set_desc_status(ps->rx_ring[index], LN_RSTATE_OWN);
    scout_flush_write_to_memory();

    ScoutAtomicSequence(--ps->rx_delayed_cnt);
} /* delayed_copyin */

#endif /* XK_SHEPHERD_PROLOG */


static int
read_frame(XObj self, int index, int len)
{
    PState *ps = (PState*) self->state;
    XShepherdBuffer *buf;
    Scout_Thread *shepherd;
    bool frame_done = TRUE;

    if (len <= sizeof(ETHhdr) PREDICT_FALSE) {
	return TRUE;			/* sanity check */
    } /* if */

    if (!self->up PREDICT_FALSE) {
	xTraceP0(self, TR_SOFT_ERROR, "no eth protocol registered");
	return TRUE;
    } /* if */

    xAssert((index >= 0) ? (ps->rx_buf[index] >= ps->ram
			    && ps->rx_buf[index] < ps->ram + 2*ps->ram_size)
	    : (ps->tx_buf[-index] >= ps->ram
	       && ps->tx_buf[-index] < ps->ram + 2*ps->ram_size));

    buf = scout_lifo_pop(&xShepherdBufferCache);
    if (buf) {
	/*
	 * Setup buffer for received message:
	 */
	buf->driverProtl = self;
	msgTruncate(&buf->upmsg, len);

#ifdef XK_SHEPHERD_PROLOG
	if (index >= 0 && ps->rx_delayed_cnt < NRX - 4 PREDICT_TRUE) {
	    xTraceP2(self, TR_FULL_TRACE,
		     "delaying copyin of frame buffer index %d (ndelayed=%d)",
		     index, ps->rx_delayed_cnt);
	    ++ps->rx_delayed_cnt;
	    buf->prolog_fun = delayed_copyin;
	    buf->prolog_arg = (void*)(long)index;
	    frame_done = FALSE;
	} else {
	    /* better get it out of LANCE quickly: */
	    copyin(buf,
		   ((index >= 0) ? ps->rx_buf[index] : ps->tx_buf[-index]),
		   len);
	} /* if */
#else
	copyin(buf, ((index >= 0) ? ps->rx_buf[index] : ps->tx_buf[-index]),
	       len);
#endif /* XK_SHEPHERD_PROLOG */

	/*
	 * Try to acquire a shepherd thread:
	 */
	shepherd = scout_lifo_pop(&xShepherdThreadCache);
	if (shepherd PREDICT_TRUE) {
	    xTraceP2(self, TR_DETAILED, "using thread `%s' (%p)",
		     scout_thread_get_name(shepherd), shepherd);
	    scout_thread_set_cont_arg(shepherd, buf);
	    scout_thread_async_wakeup(shepherd);
	} else {
	    /* No thread: append to full buffer queue: */
	    xTraceP1(self, TR_DETAILED, "no thread: appending %p to queue",
		     buf);
	    scout_fifo_append(&xShepherdFullBuffers, buf);
	} /* if */
    } else {
	xTraceP0(self, TR_SOFT_ERROR,
		 "read_frame: no buffer available: dropping incoming frame");
    } /* if */
    return frame_done;
} /* read_frame */


static inline void
rx_intr(XObj self)
{
    PState *ps = (PState*) self->state;
    int index, first, ring_cnt;
    short len;
    u_char status, status1;

    scout_set_led(SCOUT_LED_LAN_RX);

    xTrace1(lancep, TR_FULL_TRACE, "ln: rx_intr(self=%p)", self);
    /*
     * Starting from where we left off, look around the receive ring and
     * pass on all complete packets:
     */
    for (;; ps->rx_last = ++index % NRX) {
	/*
	 * Read in current descriptor:
	 */
	status = get_desc_status(ps->rx_ring[ps->rx_last]);
	if (status & LN_RSTATE_OWN) {
	    break;	/* that's it---no more work to be done */
	} /* if */
	first = index = ps->rx_last;

	/* if not the start of a packet then there was an error: */
	if (!(status & LN_RSTATE_STP) PREDICT_FALSE) {
	    xTraceP2(self, TR_ERRORS,
		     "rx_ring[%d], status=%#x, not start of packet",
		     index, status);
	    break;
	} /* if */
	/*
	 * See if packet is chained (should not) by looking at the
	 * last descriptor (OWN clear and ENP set).  Remember the
	 * status info in this last descriptor.
	 */
	if (!(status & (LN_RSTATE_ERR | LN_RSTATE_OWN | LN_RSTATE_ENP))
	    PREDICT_FALSE)
	{
	    ring_cnt = 1;
	    do {
		index = (index + 1) % NRX;
		status1 = get_desc_status(ps->rx_ring[index]);
	    } while (!(status1 & (LN_RSTATE_ERR|LN_RSTATE_OWN|LN_RSTATE_ENP))
		     && (ring_cnt++ <= NRX));
	    /*
	     * Chained packet (--> illegally sized!); re-init the
	     * descriptors involved and ignore this bogus packet.  I dunno
	     * how, but it really happens that we get these monsters.
	     */
	    xTraceP1(self, TR_ERRORS, "rx_intr: got chained pkt (cnt=%d)",
		     ring_cnt);
	    /*
	     * Return all descriptors to LANCE:
	     */
	    set_desc_status(ps->rx_ring[first], LN_RSTATE_OWN);
	    while (first != index) {
		first = (first + 1) % NRX;
		set_desc_status(ps->rx_ring[first], LN_RSTATE_OWN);
	    } /* while */
	    scout_flush_write_to_memory();
	    if (status1 & LN_RSTATE_ERR) {
		xTraceP1(self, TR_ERRORS, "rx error %#x (chained)", status1);
	    } /* if */
	    continue;
	} /* if */

	/*
	 * Good packets must be owned by us and have the end of
	 * packet flag.  And nothing else. 
	 */
	if ((status & ~LN_RSTATE_STP) == LN_RSTATE_ENP PREDICT_TRUE) {
	    len = get_desc_rx_bcnt(ps->rx_ring[ps->rx_last]);
	    if (!len PREDICT_FALSE) {
		/*
		 * When we return ownership of a read buffer to LANCE,
		 * we write a 0 in the message_size field.  Upon
		 * receiving a packet, LANCE will overwrite this with
		 * the actual length.  Apparently, LANCE sometimes
		 * hands over ownership to us, *before* setting the
		 * length.  When this happens, we re-read the
		 * descriptor.
		 */
		xTraceP0(self, TR_ERRORS,
			 "LANCE transferred ownership prematurely!");
		continue;
	    } /* if */

	    /*
	     * Drop trailing CRC bytes from len and ship packet
	     * up, minus the checksum:
	     */
	    if (read_frame(self, first, len - 4)) {
		/* copyout done already---return descriptor to LANCE: */
		set_desc_status(ps->rx_ring[first], LN_RSTATE_OWN);
		scout_flush_write_to_memory();
	    } /* if */
	} else if (status & LN_RSTATE_ERR) {
	    /*
	     * Not a good packet, see what is wrong:
	     */
	    xTraceP1(self, TR_ERRORS, "rx error (%#x)", status);

	    /*
	     * Return descriptor to LANCE:
	     */
	    set_desc_status(ps->rx_ring[first], LN_RSTATE_OWN);
	    scout_flush_write_to_memory();
	} else {
	    /* race condition with LANCE.  Wait for the next interrupt: */
	    break;
	} /* if */
    } /* for */
} /* rx_intr */


static inline void
tx_intr(XObj self)
{
    PState *ps = (PState*) self->state;
    int index;
    u_char status;
    Scout_Thread *t;
    u_char tx_status;

    xTrace1(lancep, TR_FULL_TRACE, "ln: tx_intr(self=%p)", self);

    /*
     * Free up descriptors for all packets in queue for which
     * transmission is complete.  Start from queue tail, stop at first
     * descriptor we do not OWN, or which is in an inconsistent state
     * (LANCE still working).
     */
    index = ps->tx_complete;
    while (ps->tx_count > 0) {
	status = get_desc_status(ps->tx_ring[index]);
	if (status & LN_TSTATE_OWN) {
	    break;		/* that's it---no more work to be done */
	} /* if */

	/* packet sent allright, release queue slot: */
	index = (index + 1) % NTX;
	--ps->tx_count;
	t = scout_fifo_remove(&ps->tx_queue);
	if (t PREDICT_FALSE) {
	    scout_thread_async_wakeup(t);
	} /*  if */

	if (status & (LN_TSTATE_DEF | LN_TSTATE_ONE | LN_TSTATE_MORE)) {
	    if (status & (LN_TSTATE_ONE | LN_TSTATE_MORE)) {
		xTraceP1(self, TR_SOFT_ERROR, "tx collision (status=%x)",
			 status);
	    } else {
		xTraceP0(self, TR_SOFT_ERROR, "tx deferred");
	    } /* if */
	    ++ps->collision_cnt;
	} /* if */

	if (status & LN_TSTATE_ERR PREDICT_FALSE) {
	    /*
	     * Transmission error:
	     */
	    tx_status = get_desc_tx_status(ps->tx_ring[index]);

	    xTraceP1(self, TR_ERRORS, "tx error (%#x)", tx_status);

	    if (tx_status & (LN_TXSTATE_RTRY | LN_TXSTATE_LCOL)) {
		if (tx_status & LN_TXSTATE_RTRY) {
		    xTraceP0(self, TR_ERRORS, "tx: excessive retransmissions");
		} else {
		    xTraceP0(self, TR_ERRORS, "tx: late collision");
		} /* if */
		++ps->collision_cnt;
	    } /* if */

	    /* restart chip on errors that disable the transmitter: */
	    if (tx_status & LN_TXSTATE_DISABLE) {
		xError("ln: restarting due to tx error");
		restart(self);
		index = ps->tx_complete;
		break;
	    } /* if */
	} /* if */
    } /* while */
    ps->tx_complete = index;
} /* tx_intr */


static void
interrupt(void *arg)
{
    XObj self = (XObj) arg;
    PState *ps;
    LANCE_Registers *regs;
    LANCE_Register csr0;

    xTrace0(lancep, TR_FULL_TRACE, "ln: interrupt()");

    xAssert(self && self->state);

    ps = (PState*) self->state;
    regs = ps->regs;

    csr0 = regs->rdp;			/* read status register: */
    while (csr0 & LN_CSR0_INTR) {
	/* clear pending interrupts: */
	regs->rdp = LN_CSR0_INEA | (csr0 & LN_CSR0_WTC);
	scout_flush_write_to_memory();

	/* check for errors first: */
	if (csr0 & LN_CSR0_ERR PREDICT_FALSE) {
	    if (csr0 & LN_CSR0_MISS) {
		/*
		 * Stop the chip to prevent a corrupt packet from
		 * being transmitted.  There is a known problem with
		 * missed packet errors causing corrupted data to be
		 * transmitted to the same host as was just
		 * transmitted, with a valid crc appended to the
		 * packet.  The only solution is to stop the chip,
		 * which will clear the LANCE silo, thus preventing
		 * the corrupt data from being sent.
		 */
		write_reg(regs, CSR0, LN_CSR0_STOP);

		++ps->miss_cnt;
		xIfTrace(lancep, TR_ERRORS) {
		    int lance = 0, i;
		    u_char status;

		    /* count # of buffers owned by LANCE: */
		    for (i = 0; i < NRX; i++) {
			status = get_desc_status(ps->rx_ring[i]);
			if (status & LN_RSTATE_OWN) {
			    ++lance;
			} /* if */
		    } /* for */
		    printf("ln: missed packet (%d) csr=%#x, LANCE %d, "
			   "host %d\n",
			   ps->miss_cnt, csr0 & 0xffff, lance, NRX - lance);
		} /* if */
		xError("ln: restarting due to missed packet");
		restart(self);
	    } /* if */ 
	    if (csr0 & LN_CSR0_BABL) {
		++ps->babbling_cnt;
		xTraceP1(self, TR_ERRORS, "tx timeout (%d)",
			 ps->babbling_cnt);
		if (csr0 & LN_CSR0_MERR) {
		    ++ps->memerr_cnt;
		    printf("%s: memory error (%d)\n",
			   self->instName, ps->memerr_cnt);

		    if (!(csr0 & LN_CSR0_RXON) || !(csr0 & LN_CSR0_TXON)) {
			xError("ln: restarting (rx or tx got turned off)");
			restart(self);
		    } /* if */
		} /* if */
	    } /* if */
	} else {
	    if (csr0 & LN_CSR0_RINT) {
		rx_intr(self);
	    } /* if */
	    if (csr0 & LN_CSR0_TINT) {
		tx_intr(self);
	    } /* if */
	} /* if */
	csr0 = regs->rdp;	    /* read status register: */
    } /* while */
} /* interrupt */


static xkern_return_t
open_enable(XObj self, XObj hlp, XObj hlptype, Part *part)
{
    xTraceP0(self, TR_FUNCTIONAL_TRACE, "open_enable");
    /*
     * Shepherd threads invoke xDemux() on our *protocol* (normally,
     * xDemux() should be invoked on sessions only).  xDemux() in
     * turn uses an XObj's "up" field to determine the protocol
     * to which to demux.  So all we do here is set this "up" pointer:
     */
    xSetUp(self, hlp);
    return XK_SUCCESS;
} /* open_enable */


static int
prot_control(XObj self, int op, char *buf, int len)
{
    PState *ps = (PState*) self->state;

    xTraceP1(self, TR_FULL_TRACE, "prot_control(op=%#x)", op);

    switch (op) {
      case GETMYHOST:
        if (len >= sizeof(ETHhost)) {
	    LANCE_Init_Block ib;
	    ETHhost *eh;

	    if (!ps || !ps->seated) {
		xError("ln: prot_control: protocol not initialized");
		return -1;
	    } /* if */

	    /* read init block from shared memory: */
	    bcopy_desc_to_host(L2H(ps->init_block), &ib, sizeof(ib));
	    eh = (ETHhost*) buf;
	    eh->high = ib.phys_addr_high;
	    eh->mid  = ib.phys_addr_med;
	    eh->low  = ib.phys_addr_low;
	    xTraceP1(self, TR_DETAILED, "prot_control: returning addr %s",
		     ethHostStr(eh));
	    return sizeof(ETHhost);
	} /* if */
	break;
      default:
	break;
    } /* switch */
    return -1;
} /* prot_control */


xkern_return_t
lance_init(XObj self)
{
    PState *ps = (PState*) self->state;
    const Scout_Device_Info *di;

    xTrace1(lancep, TR_FUNCTIONAL_TRACE, "lance_init(self=%p)", self);
    if (!ps || !ps->seated)  {
	ps = (PState*) xMalloc(sizeof(PState));
	bzero(ps, sizeof(*ps));
	self->state = ps;

	if (!self->instName) {
	    self->instName = "ln0";
	} /* if */

	/* find device: */
	di = scout_device_info(self->instName);
	if (!di) {
	    xTrace1(lancep, TR_DETAILED,
		    "lance_init: unknown device %s", self->instName);
	    xError("lance_init: unknown device.");
	    return XK_FAILURE;
	} /* if */

	ps->regs   = di->base_addr;
	ps->rom    = di->base_addr + di->rom_offset;
	ps->my_ipl = di->ipl;

	/* enable LANCE interrupts and set interrupt handler and arg: */
	scout_device_set_interrupt(di, 1, interrupt, self);

	if (attach(self)) {
	    self->openenable = open_enable;
	    self->push	     = lance_push;
	    self->control    = prot_control;

	    ps->seated = TRUE;
	    xTraceP0(self, TR_FUNCTIONAL_TRACE,
		     "lance_init: returning successfully");
	    return XK_SUCCESS;
	} /* if */
    } else {
	xError("lance_init: driver initialized already.");
    } /* if */
    return XK_FAILURE;
} /* lance_init */

			/*** end of lance.c ***/
