/*
 * $RCSfile: usc_validate.c,v $
 *
 * Universal Stub Compiler
 *
 * Copyright (c) 1994    Arizona Board of Regents
 *
 * $Revision: 1.1 $
 * $Date: 1994/10/13 17:43:30 $
 */

#include "usc.h"
#include "codegen.h"
#include "y.tab.h"

void init_validate (void)
{
  genout ("\n#include <stdio.h>\n");
  if (speed_test) {
    genout ("#include <time.h>\n");
  }
  genout ("\n");
  genout ("\n\nmain()\n{\n");
  genout ("  int __i;\n");
  if (speed_test) {
    genout ("  clock_t THECLOCK = clock();\n");
    genout ("  for (__i = 0; __i < 500000; __i++) {\n");
  }
}

void gen_sym_def (Symbol *sym)
{
  char type_name[128], *fmt;

  if ( TYPE_TDEF(SYM_TYPE(sym)) )
    genout ("typedef signed char %s[%d];\n", SYM_NAME(sym),SYM_MEMSIZE (sym));
  else {
    strcpy( type_name, symtype_to_string(sym) );
    if ( ( fmt = strchr(type_name,'%') ) != NULL )
      *fmt = '\0';
    genout ("volatile %s %s;\n", type_name,SYM_NAME(sym));
  }
  /*
  genout ("volatile signed char %s[%d];\n", SYM_NAME (sym), 
			 SYM_MEMSIZE (sym));
  */
}

void close_validate (void)
{
  if (speed_test) {
    genout ("  }\n");
    genout ("  printf (\"%%d microsec.\\n\", clock() - THECLOCK);\n");
  }
  genout ("}\n");
}

void gen_validate_code (char *ident, Expr *statement)
{
  Symbol *leftbase = NULL, *rightbase = NULL;
  int leftsize, rightsize;

  genout ("  /* %s */\n", ident);

  if (!EXPR_IS_ASSIGN (statement))
    crash ("Only simple assignment expressions please.");
  
  if (EXPR_IS_SYMINFO (EXPR_OP_LEFT (statement)) &&
      EXPR_IS_SYMINFO (EXPR_OP_RIGHT (statement)) ) {
    leftbase = EXPR_BASESYM (EXPR_OP_LEFT (statement));
    rightbase = EXPR_BASESYM (EXPR_OP_RIGHT (statement));
  }
  else {
    crash ("Illegal expression");
  }
  leftsize = SYM_MEMSIZE (leftbase);
  rightsize = SYM_MEMSIZE (rightbase);
  
  if (!speed_test) {
    genout ("  for (__i = 0; __i < %d; __i++) {\n", leftsize);
    genout ("    %s[__i] = 0;\n", SYM_NAME (leftbase));
    genout ("  }\n");
    genout ("  for (__i = 0; __i < %d; __i++) {\n", rightsize);
    genout ("    %s[__i] = (signed char)(__i+1);\n", SYM_NAME (rightbase));
    genout ("    printf (\"%s in[%%d]: %%02.2x\\n\", __i, %s[__i]);\n", 
            ident, SYM_NAME (rightbase));
    genout ("  }\n");
  }
  if (strcmp (SYM_NAME (leftbase), SYM_NAME (rightbase)))
    genout ("  %s((void *)%s, (void *)%s", ident, SYM_NAME (leftbase),
            SYM_NAME (rightbase));
  else
    genout ("  %s((void *)%s", ident, SYM_NAME (leftbase));
  if (!SYM_COMPLETE (leftbase))
    genout (", 0");
  genout (");\n");
  if (!speed_test) {
    genout ("  for (__i = 0; __i < %d; __i++) {\n", leftsize);
    genout ("    printf (\"%s out[%%d]: %%02.2x\\n\", __i, %s[__i]);\n", 
            ident, SYM_NAME (leftbase));
    genout ("  }\n");
    genout ("  for (__i = 0; __i < %d; __i++) {\n", leftsize);
    genout ("    %s[__i] = 0;\n", SYM_NAME (leftbase));
    genout ("  }\n");
    genout ("  for (__i = 0; __i < %d; __i++) {\n", rightsize);
    genout ("    %s[__i] = (signed char)(-__i-1);\n", SYM_NAME (rightbase));
    genout ("    printf (\"%s in[%%d]: %%02.2x\\n\", __i, %s[__i]);\n", 
            ident, SYM_NAME (rightbase));
    genout ("  }\n");
  }
  if (strcmp (SYM_NAME (leftbase), SYM_NAME (rightbase)))
    genout ("  %s((void *)%s, (void *)%s", ident, SYM_NAME (leftbase),
            SYM_NAME (rightbase));
  else
    genout ("  %s((void *)%s", ident, SYM_NAME (leftbase));
  if (!SYM_COMPLETE (leftbase))
    genout (", 0");
  genout (");\n");
  if (!speed_test) {
    genout ("  for (__i = 0; __i < %d; __i++) {\n", leftsize);
    genout ("    printf (\"%s out[%%d]: %%02.2x\\n\", __i, %s[__i]);\n", 
            ident, SYM_NAME (leftbase));
    genout ("  }\n");
    genout ("  printf (\"\\n\");\n\n");
  }
}
